@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2005, 2022, Oracle and/or its affiliates. All rights reserved.
+ * Copyright (c) 2005, 2025, Oracle and/or its affiliates. All rights reserved.
  * ORACLE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
  *
  *
@@ -54,14 +54,12 @@ import java.util.Arrays;
 import java.util.Collection;
 import java.util.Collections;
 import java.util.Comparator;
-import java.util.EnumSet;
 import java.util.HashMap;
 import java.util.Iterator;
 import java.util.Map;
 import java.util.Objects;
 import java.util.ServiceLoader;
 import java.util.Set;
-import java.util.stream.Collectors;
 import java.util.stream.Stream;
 import java.util.zip.ZipException;
 
@@ -82,6 +80,7 @@ import com.sun.tools.javac.util.DefinedBy;
 import com.sun.tools.javac.util.DefinedBy.Api;
 import com.sun.tools.javac.util.List;
 import com.sun.tools.javac.util.ListBuffer;
+import com.sun.tools.javac.util.Options;
 
 import static java.nio.charset.StandardCharsets.US_ASCII;
 import static java.nio.file.FileVisitOption.FOLLOW_LINKS;
@@ -111,7 +110,7 @@ public class JavacFileManager extends BaseFileManager implements StandardJavaFil
     private static final Set<JavaFileObject.Kind> SOURCE_OR_CLASS =
         Set.of(JavaFileObject.Kind.SOURCE, JavaFileObject.Kind.CLASS);
 
-    protected boolean symbolFileEnabled;
+    protected boolean symbolFileEnabled = true;
 
     private PathFactory pathFactory = Paths::get;
 
@@ -171,8 +170,12 @@ public class JavacFileManager extends BaseFileManager implements StandardJavaFil
     @Override
     public void setContext(Context context) {
         super.setContext(context);
-
         fsInfo = FSInfo.instance(context);
+    }
+
+    @Override
+    protected void applyOptions(Options options) {
+        super.applyOptions(options);
 
         symbolFileEnabled = !options.isSet("ignore.symbol.file");
 
@@ -193,7 +196,7 @@ public class JavacFileManager extends BaseFileManager implements StandardJavaFil
     }
 
     /**
-     * Set whether or not to use ct.sym as an alternate to rt.jar.
+     * Set whether or not to use ct.sym as an alternate to the current runtime.
      */
     public void setSymbolFileEnabled(boolean b) {
         symbolFileEnabled = b;
@@ -327,7 +330,7 @@ public class JavacFileManager extends BaseFileManager implements StandardJavaFil
             } else {
                 try {
                     fs = new ArchiveContainer(path);
-                } catch (ProviderNotFoundException | SecurityException ex) {
+                } catch (ProviderNotFoundException ex) {
                     throw new IOException(ex);
                 }
             }
@@ -556,10 +559,15 @@ public class JavacFileManager extends BaseFileManager implements StandardJavaFil
         private final FileSystem fileSystem;
         private final Map<RelativeDirectory, Path> packages;
 
-        public ArchiveContainer(Path archivePath) throws IOException, ProviderNotFoundException, SecurityException {
+        public ArchiveContainer(Path archivePath) throws IOException, ProviderNotFoundException {
             this.archivePath = archivePath;
+            Map<String,String> env = new HashMap<>();
+            // ignores timestamps not stored in ZIP central directory, reducing I/O
+            // This key is handled by ZipFileSystem only.
+            env.put("zipinfo-time", "false");
+
             if (multiReleaseValue != null && archivePath.toString().endsWith(".jar")) {
-                Map<String,String> env = Collections.singletonMap("multi-release", multiReleaseValue);
+                env.put("multi-release", multiReleaseValue);
                 FileSystemProvider jarFSProvider = fsInfo.getJarFSProvider();
                 Assert.checkNonNull(jarFSProvider, "should have been caught before!");
                 try {
@@ -568,7 +576,9 @@ public class JavacFileManager extends BaseFileManager implements StandardJavaFil
                     throw new IOException("ZipException opening \"" + archivePath.getFileName() + "\": " + ze.getMessage(), ze);
                 }
             } else {
-                this.fileSystem = FileSystems.newFileSystem(archivePath, (ClassLoader)null);
+                // Less common case is possible if the file manager was not initialized in JavacTask,
+                // or if non "*.jar" files are on the classpath.
+                this.fileSystem = FileSystems.newFileSystem(archivePath, env, (ClassLoader)null);
             }
             packages = new HashMap<>();
             for (Path root : fileSystem.getRootDirectories()) {
@@ -910,13 +920,21 @@ public class JavacFileManager extends BaseFileManager implements StandardJavaFil
             if (getClassOutDir() != null) {
                 dir = getClassOutDir();
             } else {
+                // Sibling is the associated source of the class file (e.g. x/y/Foo.java).
+                // The base name for class output is the class file name (e.g. "Foo.class").
                 String baseName = fileName.basename();
-                if (sibling != null && sibling instanceof PathFileObject pathFileObject) {
+                // Use the sibling to determine the output location where possible, unless
+                // it is in a JAR/ZIP file (we don't attempt to write class files back into
+                // archives).
+                if (sibling instanceof PathFileObject pathFileObject && !pathFileObject.isJarFile()) {
                     return pathFileObject.getSibling(baseName);
                 } else {
-                    Path p = getPath(baseName);
-                    Path real = fsInfo.getCanonicalFile(p);
-                    return PathFileObject.forSimplePath(this, real, p);
+                    // Without the sibling present, we just create an output path in the
+                    // current working directory (this isn't great, but it is what older
+                    // versions of the JDK did).
+                    Path userPath = getPath(baseName);
+                    Path realPath = fsInfo.getCanonicalFile(userPath);
+                    return PathFileObject.forSimplePath(this, realPath, userPath);
                 }
             }
         } else if (location == SOURCE_OUTPUT) {
