@@ -3,9 +3,9 @@ package com.sun.tools.jdeps.resources;
 public final class version extends java.util.ListResourceBundle {
     protected final Object[][] getContents() {
         return new Object[][] {
-            { "full", "24.0.2+11-jvmci-b01" },
-            { "jdk", "24.0.2" },
-            { "release", "24.0.2" },
+            { "full", "25.0.1+8-LTS-jvmci-b01" },
+            { "jdk", "25.0.1" },
+            { "release", "25.0.1" },
         };
     }
 }
