@@ -3,9 +3,9 @@ package jdk.javadoc.internal.tool.resources;
 public final class version extends java.util.ListResourceBundle {
     protected final Object[][] getContents() {
         return new Object[][] {
-            { "full", "21.0.10+8-LTS-jvmci-23.1-b84" },
-            { "jdk", "21.0.10" },
-            { "release", "21.0.10" },
+            { "full", "25.0.1+8-LTS-jvmci-b01" },
+            { "jdk", "25.0.1" },
+            { "release", "25.0.1" },
         };
     }
 }
