@@ -31,16 +31,6 @@ import java.util.Arrays;
 import java.util.List;
 
 import org.graalvm.collections.EconomicMap;
-import org.graalvm.compiler.api.replacements.Fold;
-import org.graalvm.compiler.core.common.NumUtil;
-import org.graalvm.compiler.core.common.SuppressFBWarnings;
-import org.graalvm.compiler.nodes.PauseNode;
-import org.graalvm.compiler.nodes.java.ArrayLengthNode;
-import org.graalvm.compiler.options.Option;
-import org.graalvm.compiler.options.OptionKey;
-import org.graalvm.compiler.options.OptionType;
-import org.graalvm.compiler.word.ObjectAccess;
-import org.graalvm.compiler.word.Word;
 import org.graalvm.nativeimage.CurrentIsolate;
 import org.graalvm.nativeimage.ImageSingletons;
 import org.graalvm.nativeimage.IsolateThread;
@@ -55,7 +45,6 @@ import org.graalvm.nativeimage.impl.ImageSingletonsSupport;
 import org.graalvm.word.Pointer;
 import org.graalvm.word.PointerBase;
 import org.graalvm.word.UnsignedWord;
-import org.graalvm.word.WordFactory;
 
 import com.oracle.svm.core.SubstrateDiagnostics.DiagnosticThunkRegistry;
 import com.oracle.svm.core.SubstrateDiagnostics.DumpCodeCacheHistory;
@@ -70,22 +59,21 @@ import com.oracle.svm.core.code.CodeInfoTable;
 import com.oracle.svm.core.code.FrameInfoQueryResult;
 import com.oracle.svm.core.code.RuntimeCodeInfoHistory;
 import com.oracle.svm.core.code.RuntimeCodeInfoMemory;
-import com.oracle.svm.core.code.UntetheredCodeInfo;
 import com.oracle.svm.core.config.ConfigurationValues;
+import com.oracle.svm.core.container.Container;
+import com.oracle.svm.core.container.OperatingSystem;
 import com.oracle.svm.core.deopt.DeoptimizationSupport;
-import com.oracle.svm.core.deopt.DeoptimizedFrame;
 import com.oracle.svm.core.deopt.Deoptimizer;
 import com.oracle.svm.core.feature.AutomaticallyRegisteredFeature;
 import com.oracle.svm.core.feature.AutomaticallyRegisteredImageSingleton;
 import com.oracle.svm.core.feature.InternalFeature;
 import com.oracle.svm.core.graal.RuntimeCompilation;
+import com.oracle.svm.core.graal.code.CGlobalDataInfo;
 import com.oracle.svm.core.graal.stackvalue.UnsafeStackValue;
 import com.oracle.svm.core.heap.Heap;
-import com.oracle.svm.core.heap.PhysicalMemory;
 import com.oracle.svm.core.heap.RestrictHeapAccess;
 import com.oracle.svm.core.hub.DynamicHub;
 import com.oracle.svm.core.hub.LayoutEncoding;
-import com.oracle.svm.core.jdk.Jvm;
 import com.oracle.svm.core.jdk.UninterruptibleUtils;
 import com.oracle.svm.core.jdk.UninterruptibleUtils.AtomicWord;
 import com.oracle.svm.core.locks.VMLockSupport;
@@ -98,8 +86,6 @@ import com.oracle.svm.core.stack.JavaFrameAnchors;
 import com.oracle.svm.core.stack.JavaStackWalker;
 import com.oracle.svm.core.stack.ThreadStackPrinter;
 import com.oracle.svm.core.stack.ThreadStackPrinter.StackFramePrintVisitor;
-import com.oracle.svm.core.stack.ThreadStackPrinter.Stage0StackFramePrintVisitor;
-import com.oracle.svm.core.stack.ThreadStackPrinter.Stage1StackFramePrintVisitor;
 import com.oracle.svm.core.thread.PlatformThreads;
 import com.oracle.svm.core.thread.VMOperation;
 import com.oracle.svm.core.thread.VMOperationControl;
@@ -113,14 +99,24 @@ import com.oracle.svm.core.util.ImageHeapList;
 import com.oracle.svm.core.util.TimeUtils;
 import com.oracle.svm.core.util.VMError;
 
+import jdk.graal.compiler.api.replacements.Fold;
+import jdk.graal.compiler.core.common.NumUtil;
+import jdk.graal.compiler.core.common.SuppressFBWarnings;
+import jdk.graal.compiler.nodes.PauseNode;
+import jdk.graal.compiler.nodes.java.ArrayLengthNode;
+import jdk.graal.compiler.options.Option;
+import jdk.graal.compiler.options.OptionKey;
+import jdk.graal.compiler.options.OptionType;
+import jdk.graal.compiler.word.ObjectAccess;
+import jdk.graal.compiler.word.Word;
+
 public class SubstrateDiagnostics {
     private static final int MAX_THREADS_TO_PRINT = 100_000;
     private static final int MAX_FRAME_ANCHORS_TO_PRINT_PER_THREAD = 1000;
 
     private static final FastThreadLocalBytes<CCharPointer> threadOnlyAttachedForCrashHandler = FastThreadLocalFactory.createBytes(() -> 1, "SubstrateDiagnostics.threadOnlyAttachedForCrashHandler");
     private static final ImageCodeLocationInfoPrinter imageCodeLocationInfoPrinter = new ImageCodeLocationInfoPrinter();
-    private static final Stage0StackFramePrintVisitor[] printVisitors = new Stage0StackFramePrintVisitor[]{new StackFramePrintVisitor(), new Stage1StackFramePrintVisitor(),
-                    new Stage0StackFramePrintVisitor()};
+    private static final StackFramePrintVisitor stackFramePrinter = new StackFramePrintVisitor();
 
     @Uninterruptible(reason = "Called from uninterruptible code.", mayBeInlined = true)
     public static void setOnlyAttachedForCrashHandler(IsolateThread thread) {
@@ -201,8 +197,8 @@ public class SubstrateDiagnostics {
         int maxPrintedBytes = elementSize == 1 ? 8 : 16;
         int printedElements = UninterruptibleUtils.Math.min(length, maxPrintedBytes / elementSize);
 
-        UnsignedWord curOffset = WordFactory.unsigned(arrayBaseOffset);
-        UnsignedWord endOffset = curOffset.add(WordFactory.unsigned(printedElements).multiply(elementSize));
+        UnsignedWord curOffset = Word.unsigned(arrayBaseOffset);
+        UnsignedWord endOffset = curOffset.add(Word.unsigned(printedElements).multiply(elementSize));
         while (curOffset.belowThan(endOffset)) {
             switch (elementSize) {
                 case 1 -> log.zhex(ObjectAccess.readByte(obj, curOffset));
@@ -224,7 +220,7 @@ public class SubstrateDiagnostics {
      * See {@link #printInformation(Log, Pointer, CodePointer, RegisterDumper.Context, boolean)}.
      */
     public static void printInformation(Log log, Pointer sp, CodePointer ip) {
-        printInformation(log, sp, ip, WordFactory.nullPointer(), false);
+        printInformation(log, sp, ip, Word.nullPointer(), false);
     }
 
     /**
@@ -257,7 +253,7 @@ public class SubstrateDiagnostics {
      * See {@link #printFatalError(Log, Pointer, CodePointer, RegisterDumper.Context, boolean)}.
      */
     public static boolean printFatalError(Log log, Pointer sp, CodePointer ip) {
-        return printFatalError(log, sp, ip, WordFactory.nullPointer(), false);
+        return printFatalError(log, sp, ip, Word.nullPointer(), false);
     }
 
     /**
@@ -380,35 +376,11 @@ public class SubstrateDiagnostics {
         log.resetIndentation().newline();
     }
 
-    @Uninterruptible(reason = "Prevent deoptimization of stack frames while in this method.")
-    private static long getTotalFrameSize(Pointer sp, CodePointer ip) {
-        DeoptimizedFrame deoptFrame = Deoptimizer.checkDeoptimized(sp);
-        if (deoptFrame != null) {
-            return deoptFrame.getSourceTotalFrameSize();
-        }
-
-        UntetheredCodeInfo untetheredInfo = CodeInfoTable.lookupCodeInfo(ip);
-        if (untetheredInfo.isNonNull()) {
-            Object tether = CodeInfoAccess.acquireTether(untetheredInfo);
-            try {
-                CodeInfo codeInfo = CodeInfoAccess.convert(untetheredInfo, tether);
-                return getTotalFrameSize0(ip, codeInfo);
-            } finally {
-                CodeInfoAccess.releaseTether(untetheredInfo, tether);
-            }
-        }
-        return -1;
-    }
-
-    @Uninterruptible(reason = "Wrap the now safe call to interruptibly look up the frame size.", calleeMustBe = false)
-    private static long getTotalFrameSize0(CodePointer ip, CodeInfo codeInfo) {
-        return CodeInfoAccess.lookupTotalFrameSize(codeInfo, CodeInfoAccess.relativeIP(codeInfo, ip));
-    }
-
     private static void logFrameAnchors(Log log, IsolateThread thread) {
         JavaFrameAnchor anchor = JavaFrameAnchors.getFrameAnchor(thread);
         if (anchor.isNull()) {
             log.string("No anchors").newline();
+            return;
         }
 
         int printed = 0;
@@ -533,7 +505,7 @@ public class SubstrateDiagnostics {
             }
             pos = pos.add(wordSize);
         }
-        return WordFactory.nullPointer();
+        return Word.nullPointer();
     }
 
     @Uninterruptible(reason = "Prevent the GC from freeing the CodeInfo.")
@@ -565,7 +537,7 @@ public class SubstrateDiagnostics {
 
         @SuppressWarnings("hiding")
         public boolean trySet(Log log, Pointer sp, CodePointer ip, RegisterDumper.Context registerContext, boolean frameHasCalleeSavedRegisters) {
-            if (diagnosticThread.compareAndSet(WordFactory.nullPointer(), CurrentIsolate.getCurrentThread())) {
+            if (diagnosticThread.compareAndSet(Word.nullPointer(), CurrentIsolate.getCurrentThread())) {
                 assert diagnosticThunkIndex == -1;
                 assert invocationCount == 0;
                 this.log = log;
@@ -584,19 +556,19 @@ public class SubstrateDiagnostics {
             log = null;
 
             ErrorContext errorContext = getErrorContext();
-            errorContext.setStackPointer(WordFactory.nullPointer());
-            errorContext.setInstructionPointer(WordFactory.nullPointer());
-            errorContext.setRegisterContext(WordFactory.nullPointer());
+            errorContext.setStackPointer(Word.nullPointer());
+            errorContext.setInstructionPointer(Word.nullPointer());
+            errorContext.setRegisterContext(Word.nullPointer());
             errorContext.setFrameHasCalleeSavedRegisters(false);
 
             diagnosticThunkIndex = -1;
             invocationCount = 0;
 
-            diagnosticThread.set(WordFactory.nullPointer());
+            diagnosticThread.set(Word.nullPointer());
         }
     }
 
-    private static class DumpRegisters extends DiagnosticThunk {
+    private static final class DumpRegisters extends DiagnosticThunk {
         @Override
         public int maxInvocationCount() {
             return 4;
@@ -618,7 +590,7 @@ public class SubstrateDiagnostics {
         }
     }
 
-    private static class DumpInstructions extends DiagnosticThunk {
+    private static final class DumpInstructions extends DiagnosticThunk {
         @Override
         public int maxInvocationCount() {
             return 4;
@@ -666,7 +638,7 @@ public class SubstrateDiagnostics {
         }
     }
 
-    private static class DumpTopOfCurrentThreadStack extends DiagnosticThunk {
+    private static final class DumpTopOfCurrentThreadStack extends DiagnosticThunk {
         @Override
         public int maxInvocationCount() {
             return 1;
@@ -723,11 +695,16 @@ public class SubstrateDiagnostics {
         @Override
         @RestrictHeapAccess(access = RestrictHeapAccess.Access.NO_ALLOCATION, reason = "Must not allocate while printing diagnostics.")
         public void printDiagnostics(Log log, ErrorContext context, int maxDiagnosticLevel, int invocationCount) {
-            log.string("DeoptStubPointer address: ").zhex(DeoptimizationSupport.getDeoptStubPointer()).newline().newline();
+            log.string("EagerDeoptStub address: ").zhex(DeoptimizationSupport.getEagerDeoptStubPointer()).newline();
+            if (Deoptimizer.Options.LazyDeoptimization.getValue()) {
+                log.string("LazyDeoptStubPrimitiveReturn address: ").zhex(DeoptimizationSupport.getLazyDeoptStubPrimitiveReturnPointer()).newline();
+                log.string("LazyDeoptStubObjectReturn address: ").zhex(DeoptimizationSupport.getLazyDeoptStubObjectReturnPointer()).newline();
+            }
+            log.newline();
         }
     }
 
-    private static class DumpThreads extends DiagnosticThunk {
+    private static final class DumpThreads extends DiagnosticThunk {
         @Override
         public int maxInvocationCount() {
             return 3;
@@ -779,7 +756,7 @@ public class SubstrateDiagnostics {
         }
     }
 
-    private static class DumpCurrentThreadLocals extends DiagnosticThunk {
+    private static final class DumpCurrentThreadLocals extends DiagnosticThunk {
         @Override
         public int maxInvocationCount() {
             return 2;
@@ -803,7 +780,7 @@ public class SubstrateDiagnostics {
         }
     }
 
-    private static class DumpCurrentVMOperation extends DiagnosticThunk {
+    private static final class DumpCurrentVMOperation extends DiagnosticThunk {
         @Override
         public int maxInvocationCount() {
             return 2;
@@ -818,7 +795,7 @@ public class SubstrateDiagnostics {
         }
     }
 
-    private static class DumpVMOperationHistory extends DiagnosticThunk {
+    private static final class DumpVMOperationHistory extends DiagnosticThunk {
         @Override
         public int maxInvocationCount() {
             return 2;
@@ -895,7 +872,7 @@ public class SubstrateDiagnostics {
             if (RuntimeCompilation.isEnabled()) {
                 log.string("Supports isolated compilation: ").bool(SubstrateOptions.supportCompileInIsolates()).newline();
             }
-            log.string("Container support: ").bool(Containers.Options.UseContainerSupport.getValue()).newline();
+            log.string("Container support: ").bool(Container.isSupported()).newline();
             log.string("Object reference size: ").signed(ConfigurationValues.getObjectLayout().getReferenceSize()).newline();
             log.string("CPU features used for AOT compiled code: ").string(getBuildTimeCpuFeatures()).newline();
             log.indent(false);
@@ -917,50 +894,112 @@ public class SubstrateDiagnostics {
         @RestrictHeapAccess(access = RestrictHeapAccess.Access.NO_ALLOCATION, reason = "Must not allocate while printing diagnostics.")
         public void printDiagnostics(Log log, ErrorContext context, int maxDiagnosticLevel, int invocationCount) {
             log.string("Runtime information:").indent(true);
-            if (SubstrateOptions.SpawnIsolates.getValue()) {
-                log.string("Isolate id: ").signed(Isolates.getIsolateId());
-                if (RuntimeCompilation.isEnabled() && IsolateArgumentParser.isCompilationIsolate()) {
-                    log.string(" (compilation isolate)");
-                }
-                log.newline();
+            log.string("Isolate id: ").signed(Isolates.getIsolateId());
+            if (RuntimeCompilation.isEnabled() && IsolateArgumentParser.isCompilationIsolate()) {
+                log.string(" (compilation isolate)");
+            }
+            log.newline();
 
-                log.string("Heap base: ").zhex(KnownIntrinsics.heapBase()).newline();
+            log.string("Heap base: ").zhex(KnownIntrinsics.heapBase()).newline();
+            if (SubstrateOptions.RelativeCodePointers.getValue()) {
+                log.string("Code base: ").zhex(KnownIntrinsics.codeBase()).newline();
+            }
+            log.string("CGlobalData base: ").zhex(CGlobalDataInfo.CGLOBALDATA_RUNTIME_BASE_ADDRESS.getPointer()).newline();
+            log.string("Containerized: ").bool(Container.singleton().isContainerized()).newline();
+
+            if (Container.singleton().isContainerized()) {
+                log.string("CPU cores (container): ");
+                int processorCount = getContainerActiveProcessorCount();
+                if (processorCount > 0) {
+                    log.signed(processorCount).newline();
+                } else {
+                    log.string("unknown").newline();
+                }
             }
 
             log.string("CPU cores (OS): ");
-            if (!SubstrateOptions.AsyncSignalSafeDiagnostics.getValue() && SubstrateOptions.JNI.getValue()) {
-                log.signed(Jvm.JVM_ActiveProcessorCount()).newline();
+            int processorCount = getOsProcessorCount();
+            if (processorCount > 0) {
+                log.signed(processorCount).newline();
             } else {
                 log.string("unknown").newline();
             }
 
-            log.string("Memory: ");
-            if (PhysicalMemory.isInitialized()) {
-                log.rational(PhysicalMemory.getCachedSize(), 1024 * 1024, 0).string("M").newline();
+            if (Container.singleton().isContainerized()) {
+                log.string("Memory (container): ");
+                UnsignedWord memory = getContainerPhysicalMemory();
+                if (memory.aboveThan(0)) {
+                    log.rational(memory, 1024 * 1024, 0).string("M").newline();
+                } else {
+                    log.string("unknown").newline();
+                }
+            }
+
+            log.string("Memory (OS): ");
+            UnsignedWord memory = getOsPhysicalMemorySize();
+            if (memory.aboveThan(0)) {
+                log.rational(memory, 1024 * 1024, 0).string("M").newline();
             } else {
                 log.string("unknown").newline();
             }
 
             log.string("Page size: ").unsigned(VirtualMemoryProvider.get().getGranularity()).newline();
             if (!SubstrateOptions.AsyncSignalSafeDiagnostics.getValue()) {
-                log.string("VM uptime: ").rational(Isolates.getCurrentUptimeMillis(), TimeUtils.millisPerSecond, 3).string("s").newline();
-                log.string("Current timestamp: ").unsigned(System.currentTimeMillis()).newline();
+                log.string("VM uptime: ").rational(Isolates.getUptimeMillis(), TimeUtils.millisPerSecond, 3).string("s").newline();
+                log.string("Current timestamp: ").unsigned(TimeUtils.currentTimeMillis()).newline();
             }
 
-            CodeInfo info = CodeInfoTable.getImageCodeInfo();
-            Pointer codeStart = (Pointer) CodeInfoAccess.getCodeStart(info);
-            UnsignedWord codeSize = CodeInfoAccess.getCodeSize(info);
-            Pointer codeEnd = codeStart.add(codeSize).subtract(1);
-            log.string("AOT compiled code: ").zhex(codeStart).string(" - ").zhex(codeEnd).newline();
+            CodeInfo info = CodeInfoTable.getFirstImageCodeInfo();
+            do {
+                Pointer codeStart = (Pointer) CodeInfoAccess.getCodeStart(info);
+                UnsignedWord codeSize = CodeInfoAccess.getCodeSize(info);
+                Pointer codeEnd = codeStart.add(codeSize).subtract(1);
+                log.string("AOT compiled code: ").zhex(codeStart).string(" - ").zhex(codeEnd).newline();
+                info = CodeInfoAccess.getNextImageCodeInfo(info);
+            } while (info.isNonNull());
 
             if (RuntimeCompilation.isEnabled()) {
                 log.string("Compile in isolates: ").bool(SubstrateOptions.shouldCompileInIsolates()).newline();
             }
             log.indent(false);
         }
+
+        private static int getOsProcessorCount() {
+            if (SubstrateOptions.AsyncSignalSafeDiagnostics.getValue()) {
+                return OperatingSystem.singleton().getCachedActiveProcessorCount();
+            } else if (SubstrateOptions.JNI.getValue()) {
+                return OperatingSystem.singleton().getActiveProcessorCount();
+            } else {
+                return -1;
+            }
+        }
+
+        private static UnsignedWord getOsPhysicalMemorySize() {
+            if (SubstrateOptions.AsyncSignalSafeDiagnostics.getValue()) {
+                return OperatingSystem.singleton().getCachedPhysicalMemorySize();
+            } else {
+                return OperatingSystem.singleton().getPhysicalMemorySize();
+            }
+        }
+
+        private static int getContainerActiveProcessorCount() {
+            if (SubstrateOptions.AsyncSignalSafeDiagnostics.getValue()) {
+                return Container.singleton().getCachedActiveProcessorCount();
+            } else {
+                return Container.singleton().getActiveProcessorCount();
+            }
+        }
+
+        private static UnsignedWord getContainerPhysicalMemory() {
+            if (SubstrateOptions.AsyncSignalSafeDiagnostics.getValue()) {
+                return Container.singleton().getCachedPhysicalMemory();
+            } else {
+                return Container.singleton().getPhysicalMemory();
+            }
+        }
     }
 
-    private static class DumpCounters extends DiagnosticThunk {
+    private static final class DumpCounters extends DiagnosticThunk {
         @Override
         public int maxInvocationCount() {
             return 1;
@@ -978,7 +1017,7 @@ public class SubstrateDiagnostics {
         }
     }
 
-    private static class DumpCurrentThreadFrameAnchors extends DiagnosticThunk {
+    private static final class DumpCurrentThreadFrameAnchors extends DiagnosticThunk {
         @Override
         public int maxInvocationCount() {
             return 1;
@@ -994,7 +1033,7 @@ public class SubstrateDiagnostics {
         }
     }
 
-    private static class DumpCurrentThreadDecodedStackTrace extends DiagnosticThunk {
+    private static final class DumpCurrentThreadDecodedStackTrace extends DiagnosticThunk {
         @Override
         public int maxInvocationCount() {
             return 3;
@@ -1006,8 +1045,8 @@ public class SubstrateDiagnostics {
             Pointer sp = context.getStackPointer();
             CodePointer ip = context.getInstructionPointer();
 
-            log.string("Stacktrace for the failing thread ").zhex(CurrentIsolate.getCurrentThread()).string(" (A=AOT compiled, J=JIT compiled, D=deoptimized, i=inlined):").indent(true);
-            boolean success = ThreadStackPrinter.printStacktrace(sp, ip, printVisitors[invocationCount - 1].reset(), log);
+            log.string("Stacktrace for the failing thread ").zhex(CurrentIsolate.getCurrentThread()).string(" (A=AOT compiled, J=JIT compiled, D=deoptimized, i=inlined, C=native):").indent(true);
+            boolean success = ThreadStackPrinter.printStacktrace(CurrentIsolate.getCurrentThread(), sp, ip, stackFramePrinter.reset(invocationCount), log);
 
             if (!success && DiagnosticLevel.unsafeOperationsAllowed(maxDiagnosticLevel)) {
                 /*
@@ -1038,11 +1077,11 @@ public class SubstrateDiagnostics {
             Pointer sp = returnAddressPos.add(FrameAccess.returnAddressSize());
             log.newline();
             log.string("Starting the stack walk in a possible caller (sp + ").unsigned(sp.subtract(originalSp)).string("):").newline();
-            ThreadStackPrinter.printStacktrace(sp, possibleIp, printVisitors[invocationCount - 1].reset(), log);
+            ThreadStackPrinter.printStacktrace(CurrentIsolate.getCurrentThread(), sp, possibleIp, stackFramePrinter.reset(invocationCount), log);
         }
     }
 
-    private static class DumpOtherStackTraces extends DiagnosticThunk {
+    private static final class DumpOtherStackTraces extends DiagnosticThunk {
         @Override
         public int maxInvocationCount() {
             return 3;
@@ -1085,12 +1124,12 @@ public class SubstrateDiagnostics {
 
         private static void printStackTrace(Log log, IsolateThread vmThread, int invocationCount) {
             log.string("Stacktrace:").indent(true);
-            JavaStackWalker.walkThread(vmThread, printVisitors[invocationCount - 1].reset(), log);
+            JavaStackWalker.walkThread(vmThread, stackFramePrinter.reset(invocationCount), log);
             log.redent(false);
         }
     }
 
-    private static class DumpCommandLine extends DiagnosticThunk {
+    private static final class DumpCommandLine extends DiagnosticThunk {
         @Override
         public int maxInvocationCount() {
             return 1;
@@ -1110,7 +1149,7 @@ public class SubstrateDiagnostics {
         }
     }
 
-    private static class ImageCodeLocationInfoPrinter {
+    private static final class ImageCodeLocationInfoPrinter {
         private final CodeInfoDecoder.FrameInfoCursor frameInfoCursor = new CodeInfoDecoder.FrameInfoCursor();
 
         /**
@@ -1120,25 +1159,33 @@ public class SubstrateDiagnostics {
          * NOTE: this method may only be called by a single thread.
          */
         public boolean printLocationInfo(Log log, UnsignedWord value) {
-            CodeInfo imageCodeInfo = CodeInfoTable.getImageCodeInfo();
-            if (imageCodeInfo.equal(value)) {
-                log.string("is the image CodeInfo object");
+            if (SubstrateOptions.RelativeCodePointers.getValue() && KnownIntrinsics.codeBase().equal(value)) {
+                log.string("is the code base");
                 return true;
             }
 
-            UnsignedWord codeInfoEnd = ((UnsignedWord) imageCodeInfo).add(CodeInfoAccess.getSizeOfCodeInfo());
-            if (value.aboveOrEqual((UnsignedWord) imageCodeInfo) && value.belowThan(codeInfoEnd)) {
-                log.string("points inside the image CodeInfo object ").zhex(imageCodeInfo);
-                return true;
-            }
-
-            if (CodeInfoAccess.contains(imageCodeInfo, (CodePointer) value)) {
-                log.string("points into AOT compiled code ");
-                FrameInfoQueryResult compilationRoot = getCompilationRoot(imageCodeInfo, (CodePointer) value);
-                if (compilationRoot != null) {
-                    compilationRoot.log(log);
+            CodeInfo imageCodeInfo = CodeInfoTable.getFirstImageCodeInfo();
+            while (imageCodeInfo.isNonNull()) {
+                if (imageCodeInfo.equal(value)) {
+                    log.string("is an image CodeInfo object");
+                    return true;
                 }
-                return true;
+
+                UnsignedWord codeInfoEnd = ((UnsignedWord) imageCodeInfo).add(CodeInfoAccess.getSizeOfCodeInfo());
+                if (value.aboveOrEqual((UnsignedWord) imageCodeInfo) && value.belowThan(codeInfoEnd)) {
+                    log.string("points inside the image CodeInfo object ").zhex(imageCodeInfo);
+                    return true;
+                }
+
+                if (CodeInfoAccess.contains(imageCodeInfo, (CodePointer) value)) {
+                    log.string("points into AOT compiled code ");
+                    FrameInfoQueryResult compilationRoot = getCompilationRoot(imageCodeInfo, (CodePointer) value);
+                    if (compilationRoot != null) {
+                        compilationRoot.log(log);
+                    }
+                    return true;
+                }
+                imageCodeInfo = CodeInfoAccess.getNextImageCodeInfo(imageCodeInfo);
             }
 
             return false;
@@ -1263,21 +1310,23 @@ public class SubstrateDiagnostics {
             if (ImageSingletons.contains(JavaMainWrapper.JavaMainSupport.class)) {
                 thunks.add(new DumpCommandLine());
             }
-            thunks.add(new DumpCounters());
+            if (CounterSupport.isEnabled()) {
+                thunks.add(new DumpCounters());
+            }
 
             resizeInitialInvocationCount();
             this.runtimeCompilationPosition = thunks.size();
         }
 
         @Platforms(Platform.HOSTED_ONLY.class)
-        public synchronized void register(DiagnosticThunk diagnosticThunk) {
+        public synchronized void add(DiagnosticThunk thunk) {
             assert !BuildPhaseProvider.isAnalysisStarted();
-            thunks.add(diagnosticThunk);
+            thunks.add(thunk);
             resizeInitialInvocationCount();
         }
 
         @Platforms(Platform.HOSTED_ONLY.class)
-        public synchronized void register(int insertPos, DiagnosticThunk... extraThunks) {
+        public synchronized void add(int insertPos, DiagnosticThunk... extraThunks) {
             assert !BuildPhaseProvider.isAnalysisStarted();
             for (int i = 0; i < extraThunks.length; i++) {
                 thunks.add(insertPos + i, extraThunks[i]);
@@ -1285,12 +1334,31 @@ public class SubstrateDiagnostics {
             resizeInitialInvocationCount();
         }
 
+        @Platforms(Platform.HOSTED_ONLY.class)
+        public synchronized void addAfter(DiagnosticThunk thunk, Class<? extends DiagnosticThunk> before) {
+            assert !BuildPhaseProvider.isAnalysisStarted();
+            int insertPos = indexOf(before) + 1;
+            assert insertPos > 0;
+            thunks.add(insertPos, thunk);
+            resizeInitialInvocationCount();
+        }
+
         @Platforms(Platform.HOSTED_ONLY.class)
         private void resizeInitialInvocationCount() {
             initialInvocationCount = new int[thunks.size()];
             Arrays.fill(initialInvocationCount, 1);
         }
 
+        @Platforms(Platform.HOSTED_ONLY.class)
+        private int indexOf(Class<? extends DiagnosticThunk> clazz) {
+            for (int i = 0; i < thunks.size(); i++) {
+                if (thunks.get(i).getClass() == clazz) {
+                    return i;
+                }
+            }
+            throw VMError.shouldNotReachHere("Could not find diagnostic thunk " + clazz);
+        }
+
         @Fold
         int size() {
             return thunks.size();
@@ -1386,7 +1454,7 @@ class SubstrateDiagnosticsFeature implements InternalFeature {
         if (RuntimeCompilation.isEnabled()) {
             int pos = registry.runtimeCompilationPosition;
             SubstrateDiagnostics.DiagnosticThunk[] thunks = {new DumpCodeCacheHistory(), new DumpRuntimeCodeInfoMemory(), new DumpDeoptStubPointer(), new DumpRecentDeoptimizations()};
-            registry.register(pos, thunks);
+            registry.add(pos, thunks);
         }
     }
 }
