@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2019, 2020, Oracle and/or its affiliates. All rights reserved.
+ * Copyright (c) 2019, 2024, Oracle and/or its affiliates. All rights reserved.
  * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
  *
  * This code is free software; you can redistribute it and/or modify it
@@ -28,18 +28,13 @@ import java.lang.reflect.Array;
 import java.nio.ByteBuffer;
 import java.util.Arrays;
 
-import org.graalvm.compiler.nodes.java.ArrayLengthNode;
-import org.graalvm.compiler.word.Word;
-import org.graalvm.nativeimage.ImageSingletons;
 import org.graalvm.nativeimage.Platform;
 import org.graalvm.nativeimage.Platforms;
 import org.graalvm.nativeimage.c.type.CTypeConversion;
-import org.graalvm.nativeimage.impl.UnmanagedMemorySupport;
 import org.graalvm.word.Pointer;
 import org.graalvm.word.PointerBase;
 import org.graalvm.word.UnsignedWord;
 import org.graalvm.word.WordBase;
-import org.graalvm.word.WordFactory;
 
 import com.oracle.svm.core.JavaMemoryUtil;
 import com.oracle.svm.core.SubstrateUtil;
@@ -53,9 +48,15 @@ import com.oracle.svm.core.heap.ReferenceAccess;
 import com.oracle.svm.core.hub.DynamicHub;
 import com.oracle.svm.core.hub.LayoutEncoding;
 import com.oracle.svm.core.jdk.UninterruptibleUtils;
+import com.oracle.svm.core.memory.NullableNativeMemory;
+import com.oracle.svm.core.nmt.NmtCategory;
 import com.oracle.svm.core.snippets.KnownIntrinsics;
 import com.oracle.svm.core.util.VMError;
 
+import jdk.graal.compiler.nodes.NamedLocationIdentity;
+import jdk.graal.compiler.nodes.java.ArrayLengthNode;
+import jdk.graal.compiler.word.Word;
+
 /**
  * Support for allocating and accessing non-moving arrays. Such arrays are safe to access during
  * garbage collection. They can also be passed between isolates, provided that they do not contain
@@ -83,7 +84,7 @@ public final class NonmovableArrays {
 
     @SuppressWarnings("unchecked")
     @Uninterruptible(reason = "Called from uninterruptible code.", mayBeInlined = true)
-    private static <T extends NonmovableArray<?>> T createArray(int length, Class<?> arrayType) {
+    private static <T extends NonmovableArray<?>> T createArray(int length, Class<?> arrayType, NmtCategory nmtCategory) {
         if (SubstrateUtil.HOSTED) {
             Class<?> componentType = arrayType.getComponentType();
             Object array = Array.newInstance(componentType, length);
@@ -92,16 +93,15 @@ public final class NonmovableArrays {
         DynamicHub hub = SubstrateUtil.cast(arrayType, DynamicHub.class);
         assert LayoutEncoding.isArray(hub.getLayoutEncoding());
         UnsignedWord size = LayoutEncoding.getArrayAllocationSize(hub.getLayoutEncoding(), length);
-        Pointer array = ImageSingletons.lookup(UnmanagedMemorySupport.class).calloc(size);
+        Pointer array = NullableNativeMemory.calloc(size, nmtCategory);
         if (array.isNull()) {
             throw OUT_OF_MEMORY_ERROR;
         }
 
         ObjectHeader header = Heap.getHeap().getObjectHeader();
         Word encodedHeader = header.encodeAsUnmanagedObjectHeader(hub);
-        header.initializeHeaderOfNewObject(array, encodedHeader);
-
-        array.writeInt(ConfigurationValues.getObjectLayout().getArrayLengthOffset(), length);
+        header.initializeHeaderOfNewObjectOffHeap(array, encodedHeader, true);
+        array.writeInt(ConfigurationValues.getObjectLayout().getArrayLengthOffset(), length, NamedLocationIdentity.OFF_HEAP_LOCATION);
         // already zero-initialized thanks to calloc()
         trackUnmanagedArray((NonmovableArray<?>) array);
         return (T) array;
@@ -147,7 +147,7 @@ public final class NonmovableArrays {
     /** Provides the size of the given array in bytes. */
     @Uninterruptible(reason = "Called from uninterruptible code.", mayBeInlined = true)
     public static UnsignedWord byteSizeOf(NonmovableArray<?> array) {
-        return array.isNonNull() ? LayoutEncoding.getArrayAllocationSize(readLayoutEncoding(array), lengthOf(array)) : WordFactory.zero();
+        return array.isNonNull() ? LayoutEncoding.getArrayAllocationSize(readLayoutEncoding(array), lengthOf(array)) : Word.zero();
     }
 
     /** @see System#arraycopy */
@@ -159,7 +159,7 @@ public final class NonmovableArrays {
         }
         assert srcPos >= 0 && destPos >= 0 && length >= 0 && srcPos + length <= lengthOf(src) && destPos + length <= lengthOf(dest);
         assert readHub(src) == readHub(dest) : "copying is only allowed with same component types";
-        UnmanagedMemoryUtil.copy(addressOf(src, srcPos), addressOf(dest, destPos), WordFactory.unsigned(length << readElementShift(dest)));
+        UnmanagedMemoryUtil.copy(addressOf(src, srcPos), addressOf(dest, destPos), Word.unsigned(length << readElementShift(dest)));
     }
 
     /** Provides an array for which {@link NonmovableArray#isNull()} returns {@code true}. */
@@ -169,7 +169,7 @@ public final class NonmovableArrays {
         if (SubstrateUtil.HOSTED) {
             return (T) HOSTED_NULL_VALUE;
         }
-        return WordFactory.nullPointer();
+        return Word.nullPointer();
     }
 
     /**
@@ -180,8 +180,8 @@ public final class NonmovableArrays {
      * array, it is not a Java object and must never be referenced as an object.
      */
     @Uninterruptible(reason = "Called from uninterruptible code.", mayBeInlined = true)
-    public static NonmovableArray<Byte> createByteArray(int nbytes) {
-        return createArray(nbytes, byte[].class);
+    public static NonmovableArray<Byte> createByteArray(int nbytes, NmtCategory nmtCategory) {
+        return createArray(nbytes, byte[].class, nmtCategory);
     }
 
     /**
@@ -192,8 +192,8 @@ public final class NonmovableArrays {
      * of a Java array, it is not a Java object and must never be referenced as an object.
      */
     @Uninterruptible(reason = "Called from uninterruptible code.", mayBeInlined = true)
-    public static NonmovableArray<Integer> createIntArray(int length) {
-        return createArray(length, int[].class);
+    public static NonmovableArray<Integer> createIntArray(int length, NmtCategory nmtCategory) {
+        return createArray(length, int[].class, nmtCategory);
     }
 
     /**
@@ -204,8 +204,8 @@ public final class NonmovableArrays {
      * of a Java array, it is not a Java object and must never be referenced as an object.
      */
     @Uninterruptible(reason = "Called from uninterruptible code.", mayBeInlined = true)
-    public static <T extends WordBase> NonmovableArray<T> createWordArray(int length) {
-        return createArray(length, WordBase[].class);
+    public static <T extends WordBase> NonmovableArray<T> createWordArray(int length, NmtCategory nmtCategory) {
+        return createArray(length, WordBase[].class, nmtCategory);
     }
 
     /**
@@ -219,16 +219,16 @@ public final class NonmovableArrays {
      * that of a Java array, it is not a Java object and must never be referenced as an object.
      */
     @Uninterruptible(reason = "Called from uninterruptible code.", mayBeInlined = true)
-    public static <T> NonmovableObjectArray<T> createObjectArray(Class<T[]> arrayType, int length) {
+    public static <T> NonmovableObjectArray<T> createObjectArray(Class<T[]> arrayType, int length, NmtCategory nmtCategory) {
         assert (SubstrateUtil.HOSTED ? (arrayType.isArray() && !arrayType.getComponentType().isPrimitive())
                         : LayoutEncoding.isObjectArray(SubstrateUtil.cast(arrayType, DynamicHub.class).getLayoutEncoding())) : "must be an object array type";
-        return createArray(length, arrayType);
+        return createArray(length, arrayType, nmtCategory);
     }
 
     /** @see java.util.Arrays#copyOf */
     @Uninterruptible(reason = "Called from uninterruptible code.", mayBeInlined = true)
-    public static <T> NonmovableObjectArray<T> copyOfObjectArray(T[] source, int newLength) {
-        NonmovableObjectArray<T> array = createArray(newLength, source.getClass());
+    public static <T> NonmovableObjectArray<T> copyOfObjectArray(T[] source, int newLength, NmtCategory nmtCategory) {
+        NonmovableObjectArray<T> array = createArray(newLength, source.getClass(), nmtCategory);
         int copyLength = (source.length < newLength) ? source.length : newLength;
         for (int i = 0; i < copyLength; i++) {
             setObject(array, i, source[i]);
@@ -238,8 +238,8 @@ public final class NonmovableArrays {
 
     /** Same as {@link #copyOfObjectArray} with a {@code newLength} of the array length. */
     @Uninterruptible(reason = "Called from uninterruptible code.", mayBeInlined = true)
-    public static <T> NonmovableObjectArray<T> copyOfObjectArray(T[] source) {
-        return copyOfObjectArray(source, source.length);
+    public static <T> NonmovableObjectArray<T> copyOfObjectArray(T[] source, NmtCategory nmtCategory) {
+        return copyOfObjectArray(source, source.length, nmtCategory);
     }
 
     public static byte[] heapCopyOfByteArray(NonmovableArray<Byte> source) {
@@ -285,7 +285,7 @@ public final class NonmovableArrays {
         Pointer destAddressAtPos = Word.objectToUntrackedPointer(dest).add(LayoutEncoding.getArrayElementOffset(destHub.getLayoutEncoding(), destPos));
         if (LayoutEncoding.isPrimitiveArray(destHub.getLayoutEncoding())) {
             Pointer srcAddressAtPos = addressOf(src, srcPos);
-            JavaMemoryUtil.copyPrimitiveArrayForward(srcAddressAtPos, destAddressAtPos, WordFactory.unsigned(length << readElementShift(src)));
+            JavaMemoryUtil.copyPrimitiveArrayForward(srcAddressAtPos, destAddressAtPos, Word.unsigned(length << readElementShift(src)));
         } else { // needs barriers
             Object[] destArr = (Object[]) dest;
             for (int i = 0; i < length; i++) {
@@ -298,7 +298,7 @@ public final class NonmovableArrays {
     /** Releases an array created at runtime. */
     @Uninterruptible(reason = "Called from uninterruptible code.", mayBeInlined = true)
     public static void releaseUnmanagedArray(NonmovableArray<?> array) {
-        ImageSingletons.lookup(UnmanagedMemorySupport.class).free(array);
+        NullableNativeMemory.free(array);
         untrackUnmanagedArray(array);
     }
 
@@ -321,7 +321,7 @@ public final class NonmovableArrays {
             return new HostedNonmovableArray<>(array);
         }
         assert array == null || Heap.getHeap().isInImageHeap(array);
-        return (array != null) ? (NonmovableArray<T>) Word.objectToUntrackedPointer(array) : WordFactory.nullPointer();
+        return (array != null) ? (NonmovableArray<T>) Word.objectToUntrackedPointer(array) : Word.nullPointer();
     }
 
     /** Returns a {@link NonmovableObjectArray} for an object array in the image heap. */
@@ -444,36 +444,29 @@ public final class NonmovableArrays {
      * Visits all array elements with the provided {@link ObjectReferenceVisitor}.
      */
     @Uninterruptible(reason = "Called from uninterruptible code.", mayBeInlined = true)
-    public static boolean walkUnmanagedObjectArray(NonmovableObjectArray<?> array, ObjectReferenceVisitor visitor) {
+    public static void walkUnmanagedObjectArray(NonmovableObjectArray<?> array, ObjectReferenceVisitor visitor) {
         if (array.isNonNull()) {
-            return walkUnmanagedObjectArray(array, visitor, 0, lengthOf(array));
+            walkUnmanagedObjectArray(array, visitor, 0, lengthOf(array));
         }
-        return true;
     }
 
     /**
      * Visits all array elements with the provided {@link ObjectReferenceVisitor}.
      */
     @Uninterruptible(reason = "Called from uninterruptible code.", mayBeInlined = true)
-    public static boolean walkUnmanagedObjectArray(NonmovableObjectArray<?> array, ObjectReferenceVisitor visitor, int startIndex, int count) {
+    public static void walkUnmanagedObjectArray(NonmovableObjectArray<?> array, ObjectReferenceVisitor visitor, int startIndex, int count) {
         if (array.isNonNull()) {
             assert startIndex >= 0 && count <= lengthOf(array) - startIndex;
             int refSize = ConfigurationValues.getObjectLayout().getReferenceSize();
             assert refSize == (1 << readElementShift(array));
-            Pointer p = ((Pointer) array).add(readArrayBase(array)).add(startIndex * refSize);
-            for (int i = 0; i < count; i++) {
-                if (!callVisitor(visitor, p)) {
-                    return false;
-                }
-                p = p.add(refSize);
-            }
+            Pointer firstObjRef = ((Pointer) array).add(readArrayBase(array)).add(startIndex * refSize);
+            callVisitor(visitor, firstObjRef, refSize, count);
         }
-        return true;
     }
 
     @Uninterruptible(reason = "Bridge between uninterruptible and potentially interruptible code.", mayBeInlined = true, calleeMustBe = false)
-    private static boolean callVisitor(ObjectReferenceVisitor visitor, Pointer p) {
-        return visitor.visitObjectReference(p, true, null);
+    private static void callVisitor(ObjectReferenceVisitor visitor, Pointer firstObjRef, int referenceSize, int count) {
+        visitor.visitObjectReferences(firstObjRef, true, referenceSize, null, count);
     }
 
     @Uninterruptible(reason = "Called from uninterruptible code.", mayBeInlined = true)
