@@ -24,7 +24,6 @@
  */
 package com.oracle.svm.core.genscavenge;
 
-import jdk.graal.compiler.api.replacements.Fold;
 import org.graalvm.nativeimage.ImageSingletons;
 import org.graalvm.word.Pointer;
 
@@ -32,6 +31,8 @@ import com.oracle.svm.core.MemoryWalker;
 import com.oracle.svm.core.Uninterruptible;
 import com.oracle.svm.core.heap.ObjectVisitor;
 
+import jdk.graal.compiler.api.replacements.Fold;
+
 public interface AuxiliaryImageHeap {
     @Fold
     static boolean isPresent() {
@@ -46,10 +47,12 @@ public interface AuxiliaryImageHeap {
     @Uninterruptible(reason = "Called from uninterruptible code.", mayBeInlined = true)
     boolean containsObject(Pointer address);
 
-    boolean walkObjects(ObjectVisitor visitor);
+    void walkObjects(ObjectVisitor visitor);
+
+    void walkHeapChunks(HeapChunkVisitor visitor);
 
     @Uninterruptible(reason = "Called from uninterruptible code.", mayBeInlined = true)
-    boolean walkRegions(MemoryWalker.ImageHeapRegionVisitor visitor);
+    void walkRegions(MemoryWalker.ImageHeapRegionVisitor visitor);
 
     ImageHeapInfo getImageHeapInfo();
 }
