@@ -24,6 +24,7 @@
  */
 package com.oracle.svm.core.genscavenge;
 
+import java.io.Serial;
 import java.lang.ref.Reference;
 import java.util.ArrayList;
 import java.util.List;
@@ -34,7 +35,6 @@ import org.graalvm.nativeimage.Platform;
 import org.graalvm.nativeimage.Platforms;
 import org.graalvm.word.Pointer;
 import org.graalvm.word.UnsignedWord;
-import org.graalvm.word.WordFactory;
 
 import com.oracle.svm.core.AlwaysInline;
 import com.oracle.svm.core.MemoryWalker;
@@ -49,11 +49,10 @@ import com.oracle.svm.core.SubstrateUtil;
 import com.oracle.svm.core.Uninterruptible;
 import com.oracle.svm.core.annotate.Substitute;
 import com.oracle.svm.core.annotate.TargetClass;
-import com.oracle.svm.core.config.ConfigurationValues;
 import com.oracle.svm.core.genscavenge.AlignedHeapChunk.AlignedHeader;
-import com.oracle.svm.core.genscavenge.ThreadLocalAllocation.Descriptor;
 import com.oracle.svm.core.genscavenge.UnalignedHeapChunk.UnalignedHeader;
-import com.oracle.svm.core.genscavenge.graal.ForcedSerialPostWriteBarrier;
+import com.oracle.svm.core.genscavenge.graal.nodes.FormatArrayNode;
+import com.oracle.svm.core.genscavenge.remset.RememberedSet;
 import com.oracle.svm.core.graal.snippets.SubstrateAllocationSnippets;
 import com.oracle.svm.core.heap.GC;
 import com.oracle.svm.core.heap.GCCause;
@@ -66,9 +65,10 @@ import com.oracle.svm.core.heap.ReferenceHandlerThread;
 import com.oracle.svm.core.heap.ReferenceInternals;
 import com.oracle.svm.core.heap.RestrictHeapAccess;
 import com.oracle.svm.core.heap.RuntimeCodeInfoGCSupport;
+import com.oracle.svm.core.heap.VMOperationInfos;
 import com.oracle.svm.core.hub.DynamicHub;
+import com.oracle.svm.core.hub.LayoutEncoding;
 import com.oracle.svm.core.imagelayer.ImageLayerBuildingSupport;
-import com.oracle.svm.core.jdk.UninterruptibleUtils.AtomicReference;
 import com.oracle.svm.core.jfr.JfrTicks;
 import com.oracle.svm.core.jfr.events.SystemGCEvent;
 import com.oracle.svm.core.layeredimagesingleton.MultiLayeredImageSingleton;
@@ -80,6 +80,7 @@ import com.oracle.svm.core.nodes.CFunctionPrologueNode;
 import com.oracle.svm.core.option.RuntimeOptionKey;
 import com.oracle.svm.core.os.ImageHeapProvider;
 import com.oracle.svm.core.snippets.KnownIntrinsics;
+import com.oracle.svm.core.thread.JavaVMOperation;
 import com.oracle.svm.core.thread.PlatformThreads;
 import com.oracle.svm.core.thread.ThreadStatus;
 import com.oracle.svm.core.thread.VMOperation;
@@ -94,7 +95,7 @@ import jdk.graal.compiler.api.replacements.Fold;
 import jdk.graal.compiler.core.common.NumUtil;
 import jdk.graal.compiler.core.common.SuppressFBWarnings;
 import jdk.graal.compiler.nodes.extended.MembarNode;
-import jdk.graal.compiler.nodes.memory.address.OffsetAddressNode;
+import jdk.graal.compiler.replacements.AllocationSnippets;
 import jdk.graal.compiler.word.Word;
 
 public final class HeapImpl extends Heap {
@@ -110,6 +111,9 @@ public final class HeapImpl extends Heap {
     private final GCImpl gcImpl;
     private final RuntimeCodeInfoGCSupportImpl runtimeCodeInfoGcSupport;
     private final HeapAccounting accounting = new HeapAccounting();
+    private final ImageHeapChunkLogger imageHeapChunkLogger = new ImageHeapChunkLogger();
+
+    private AlignedHeader lastDynamicHubChunk;
 
     /** Head of the linked list of currently pending (ready to be enqueued) {@link Reference}s. */
     private Reference<?> refPendingList;
@@ -118,9 +122,6 @@ public final class HeapImpl extends Heap {
     /** Total number of times when threads waiting for a pending reference list were interrupted. */
     private volatile long refListWaiterWakeUpCounter;
 
-    /** Head of the linked list of object pins. */
-    private final AtomicReference<PinnedObjectImpl> pinHead = new AtomicReference<>();
-
     /** A cached list of all the classes, if someone asks for it. */
     private List<Class<?>> classList;
 
@@ -133,7 +134,9 @@ public final class HeapImpl extends Heap {
         HeapParameters.initialize();
         DiagnosticThunkRegistry.singleton().add(new DumpHeapSettingsAndStatistics());
         DiagnosticThunkRegistry.singleton().add(new DumpHeapUsage());
-        DiagnosticThunkRegistry.singleton().add(new DumpChunkInformation());
+        DiagnosticThunkRegistry.singleton().add(new DumpGCPolicy());
+        DiagnosticThunkRegistry.singleton().add(new DumpImageHeapInfo());
+        DiagnosticThunkRegistry.singleton().add(new DumpChunkInfo());
     }
 
     @Fold
@@ -193,7 +196,7 @@ public final class HeapImpl extends Heap {
     @Override
     @Uninterruptible(reason = "Called from uninterruptible code.", mayBeInlined = true)
     public void suspendAllocation() {
-        ThreadLocalAllocation.suspendInCurrentThread();
+        TlabSupport.suspendAllocationInCurrentThread();
     }
 
     @Override
@@ -202,9 +205,10 @@ public final class HeapImpl extends Heap {
     }
 
     @Override
-    public boolean walkObjects(ObjectVisitor visitor) {
+    public void walkObjects(ObjectVisitor visitor) {
         VMOperation.guaranteeInProgressAtSafepoint("must only be executed at a safepoint");
-        return walkImageHeapObjects(visitor) && walkCollectedHeapObjects(visitor);
+        walkImageHeapObjects(visitor);
+        walkCollectedHeapObjects(visitor);
     }
 
     /** Tear down the heap and release its memory. */
@@ -259,7 +263,7 @@ public final class HeapImpl extends Heap {
     /** A guard to place before an allocation, giving the call site and the allocation type. */
     static void exitIfAllocationDisallowed(String callSite, String typeName) {
         if (HeapImpl.getHeapImpl().isAllocationDisallowed()) {
-            NoAllocationVerifier.exit(callSite, typeName);
+            throw NoAllocationVerifier.exit(callSite, typeName);
         }
     }
 
@@ -273,40 +277,24 @@ public final class HeapImpl extends Heap {
         return oldGeneration;
     }
 
-    @Fold
-    AtomicReference<PinnedObjectImpl> getPinHead() {
-        return pinHead;
-    }
-
     void logUsage(Log log) {
         youngGeneration.logUsage(log);
         oldGeneration.logUsage(log);
     }
 
-    void logChunks(Log log) {
-        getYoungGeneration().logChunks(log);
+    void logChunks(Log log, boolean allowUnsafe) {
+        imageHeapChunkLogger.initialize(log);
+        for (ImageHeapInfo info : HeapImpl.getImageHeapInfos()) {
+            ImageHeapWalker.walkImageHeapChunks(info, imageHeapChunkLogger);
+        }
+        if (AuxiliaryImageHeap.isPresent()) {
+            AuxiliaryImageHeap.singleton().walkHeapChunks(imageHeapChunkLogger);
+        }
+        getYoungGeneration().logChunks(log, allowUnsafe);
         getOldGeneration().logChunks(log);
         getChunkProvider().logFreeChunks(log);
     }
 
-    @SuppressWarnings("static-method")
-    void logImageHeapPartitionBoundaries(Log log) {
-        log.string("Native image heap boundaries:").indent(true);
-        for (ImageHeapInfo info : HeapImpl.getImageHeapInfos()) {
-            info.print(log);
-        }
-        log.indent(false);
-
-        if (AuxiliaryImageHeap.isPresent()) {
-            ImageHeapInfo auxHeapInfo = AuxiliaryImageHeap.singleton().getImageHeapInfo();
-            if (auxHeapInfo != null) {
-                log.string("Auxiliary image heap boundaries:").indent(true);
-                auxHeapInfo.print(log);
-                log.indent(false);
-            }
-        }
-    }
-
     /** Log the zap values to make it easier to search for them. */
     static void logZapValues(Log log) {
         if (HeapParameters.getZapProducedHeapChunks() || HeapParameters.getZapConsumedHeapChunks()) {
@@ -361,14 +349,28 @@ public final class HeapImpl extends Heap {
 
         ArrayList<Class<?>> list = new ArrayList<>(dynamicHubCount);
         for (ImageHeapInfo info : HeapImpl.getImageHeapInfos()) {
-            ImageHeapWalker.walkRegions(info, new ClassListBuilderVisitor(list.size() + info.dynamicHubCount, list));
+            collectDynamicHubs(info, list);
         }
 
         VMError.guarantee(dynamicHubCount == list.size(), "Found fewer DynamicHubs in the image heap than expected.");
         return list;
     }
 
+    private static void collectDynamicHubs(ImageHeapInfo info, ArrayList<Class<?>> list) {
+        try {
+            ClassListBuilderVisitor visitor = new ClassListBuilderVisitor(list.size() + info.dynamicHubCount, list);
+            ImageHeapWalker.walkRegions(info, visitor);
+        } catch (FoundAllHubsException ignored) {
+        }
+    }
+
+    /**
+     * Iterates over the image heap parts that may contain dynamic hubs. After collecting all
+     * dynamic hubs were collected, this visitor throws a {@link FoundAllHubsException}.
+     */
     private static class ClassListBuilderVisitor implements MemoryWalker.ImageHeapRegionVisitor, ObjectVisitor {
+        private static final FoundAllHubsException FOUND_ALL_HUBS_EXCEPTION = new FoundAllHubsException();
+
         private final int dynamicHubCount;
         private final List<Class<?>> list;
 
@@ -378,21 +380,32 @@ public final class HeapImpl extends Heap {
         }
 
         @Override
-        public <T> boolean visitNativeImageHeapRegion(T region, MemoryWalker.NativeImageHeapRegionAccess<T> access) {
+        public <T> void visitNativeImageHeapRegion(T region, MemoryWalker.NativeImageHeapRegionAccess<T> access) {
             if (!access.isWritable(region) && !access.consistsOfHugeObjects(region)) {
-                return access.visitObjects(region, this);
+                access.visitObjects(region, this);
             }
-            return true;
         }
 
         @Override
         @RestrictHeapAccess(access = RestrictHeapAccess.Access.UNRESTRICTED, reason = "Allocation is fine: this method traverses only the image heap.")
-        public boolean visitObject(Object o) {
+        public void visitObject(Object o) {
             if (o instanceof Class<?>) {
                 list.add((Class<?>) o);
-                return list.size() != dynamicHubCount;
+                if (list.size() == dynamicHubCount) {
+                    throw FOUND_ALL_HUBS_EXCEPTION;
+                }
             }
-            return true;
+        }
+    }
+
+    private static final class FoundAllHubsException extends RuntimeException {
+        @Serial private static final long serialVersionUID = 1L;
+
+        @Override
+        @SuppressWarnings("sync-override")
+        public Throwable fillInStackTrace() {
+            /* No stacktrace needed. */
+            return this;
         }
     }
 
@@ -409,13 +422,14 @@ public final class HeapImpl extends Heap {
     @Uninterruptible(reason = "Called during startup.")
     @Override
     public void attachThread(IsolateThread isolateThread) {
-        // nothing to do
+        TlabSupport.startupInitialization();
+        TlabSupport.initialize(isolateThread);
     }
 
     @Override
     @Uninterruptible(reason = "Thread is detaching and holds the THREAD_MUTEX.")
     public void detachThread(IsolateThread isolateThread) {
-        ThreadLocalAllocation.disableAndFlushForThread(isolateThread);
+        TlabSupport.disableAndFlushForThread(isolateThread);
     }
 
     @Fold
@@ -479,24 +493,26 @@ public final class HeapImpl extends Heap {
     }
 
     @Override
-    public boolean walkImageHeapObjects(ObjectVisitor visitor) {
+    public void walkImageHeapObjects(ObjectVisitor visitor) {
         VMOperation.guaranteeInProgressAtSafepoint("Must only be called at a safepoint");
         if (visitor == null) {
-            return true;
+            return;
         }
+
         for (ImageHeapInfo info : HeapImpl.getImageHeapInfos()) {
-            if (!ImageHeapWalker.walkImageHeapObjects(info, visitor)) {
-                return false;
-            }
+            ImageHeapWalker.walkImageHeapObjects(info, visitor);
+        }
+        if (AuxiliaryImageHeap.isPresent()) {
+            AuxiliaryImageHeap.singleton().walkObjects(visitor);
         }
-        return !AuxiliaryImageHeap.isPresent() || AuxiliaryImageHeap.singleton().walkObjects(visitor);
     }
 
     @Override
-    public boolean walkCollectedHeapObjects(ObjectVisitor visitor) {
+    public void walkCollectedHeapObjects(ObjectVisitor visitor) {
         VMOperation.guaranteeInProgressAtSafepoint("Must only be called at a safepoint");
-        ThreadLocalAllocation.disableAndFlushForAllThreads();
-        return getYoungGeneration().walkObjects(visitor) && getOldGeneration().walkObjects(visitor);
+        makeParseable();
+        getYoungGeneration().walkObjects(visitor);
+        getOldGeneration().walkObjects(visitor);
     }
 
     @Override
@@ -506,6 +522,10 @@ public final class HeapImpl extends Heap {
         }
     }
 
+    void makeParseable() {
+        youngGeneration.makeParseable();
+    }
+
     @SuppressFBWarnings(value = "VO_VOLATILE_INCREMENT", justification = "Only the GC increments the volatile field 'refListOfferCounter'.")
     void addToReferencePendingList(Reference<?> list) {
         assert VMOperation.isGCInProgress();
@@ -695,7 +715,7 @@ public final class HeapImpl extends Heap {
     @Override
     @Uninterruptible(reason = "Called from uninterruptible code.", mayBeInlined = true)
     public long getThreadAllocatedMemory(IsolateThread thread) {
-        UnsignedWord allocatedBytes = ThreadLocalAllocation.allocatedBytes.getVolatile(thread);
+        UnsignedWord allocatedBytes = ThreadLocalAllocation.getAllocatedBytes(thread);
 
         /*
          * The current aligned chunk in the TLAB is only partially filled and therefore not yet
@@ -705,9 +725,8 @@ public final class HeapImpl extends Heap {
          * can cause that the memory in the current aligned TLAB chunk is counted twice.
          */
         ThreadLocalAllocation.Descriptor tlab = ThreadLocalAllocation.getTlab(thread);
-        AlignedHeader alignedTlab = tlab.getAlignedChunk();
+        Pointer start = tlab.getAlignedAllocationStart(SubstrateAllocationSnippets.TLAB_START_IDENTITY);
         Pointer top = tlab.getAllocationTop(SubstrateAllocationSnippets.TLAB_TOP_IDENTITY);
-        Pointer start = AlignedHeapChunk.getObjectsStart(alignedTlab);
 
         if (top.aboveThan(start)) {
             UnsignedWord usedTlabSize = top.subtract(start);
@@ -721,9 +740,11 @@ public final class HeapImpl extends Heap {
     @Override
     @Uninterruptible(reason = "Ensure that no GC can occur between modification of the object and this call.", callerMustBe = true)
     public void dirtyAllReferencesOf(Object obj) {
-        if (SerialGCOptions.useRememberedSet() && obj != null) {
-            ForcedSerialPostWriteBarrier.force(OffsetAddressNode.address(obj, 0), false);
+        if (obj == null) {
+            return;
         }
+
+        RememberedSet.get().dirtyAllReferencesIfNecessary(obj);
     }
 
     @Override
@@ -773,26 +794,36 @@ public final class HeapImpl extends Heap {
         if (AuxiliaryImageHeap.isPresent() && AuxiliaryImageHeap.singleton().containsObject(ptr)) {
             log.string("points into the auxiliary image heap");
             return true;
-        } else if (printTlabInfo(log, ptr, CurrentIsolate.getCurrentThread())) {
+        } else if (TlabSupport.printTlabInfo(log, ptr, CurrentIsolate.getCurrentThread())) {
             return true;
         }
 
         if (allowJavaHeapAccess) {
             // Accessing spaces and chunks is safe if we prevent a GC.
-            if (youngGeneration.isInSpace(ptr)) {
-                log.string("points into the young generation");
+            if (youngGeneration.printLocationInfo(log, ptr)) {
                 return true;
-            } else if (oldGeneration.isInSpace(ptr)) {
-                log.string("points into the old generation");
+            } else if (oldGeneration.printLocationInfo(log, ptr)) {
                 return true;
             }
         }
 
         if (allowUnsafeOperations || VMOperation.isInProgressAtSafepoint()) {
-            // If we are not at a safepoint, then it is unsafe to access thread locals of another
-            // thread as the IsolateThread could be freed at any time.
-            return printTlabInfo(log, ptr);
+            /*
+             * If we are not at a safepoint, then it is unsafe to access thread locals of another
+             * thread as the IsolateThread could be freed at any time.
+             */
+            if (TlabSupport.printTlabInfo(log, ptr)) {
+                return true;
+            }
         }
+
+        if (allowJavaHeapAccess) {
+            // Accessing chunks is safe if we prevent a GC.
+            if (YoungGeneration.getHeapAllocation().printLocationInfo(log, ptr)) {
+                return true;
+            }
+        }
+
         return false;
     }
 
@@ -800,54 +831,12 @@ public final class HeapImpl extends Heap {
         return isInImageHeap(ptr) || youngGeneration.isInSpace(ptr) || oldGeneration.isInSpace(ptr);
     }
 
-    private static boolean printTlabInfo(Log log, Pointer ptr) {
-        for (IsolateThread thread = VMThreads.firstThreadUnsafe(); thread.isNonNull(); thread = VMThreads.nextThread(thread)) {
-            if (printTlabInfo(log, ptr, thread)) {
-                return true;
-            }
-        }
-        return false;
-    }
-
-    private static boolean printTlabInfo(Log log, Pointer ptr, IsolateThread thread) {
-        ThreadLocalAllocation.Descriptor tlab = getTlabUnsafe(thread);
-        AlignedHeader aChunk = tlab.getAlignedChunk();
-        while (aChunk.isNonNull()) {
-            Pointer dataStart = AlignedHeapChunk.getObjectsStart(aChunk);
-            Pointer dataEnd = AlignedHeapChunk.getObjectsEnd(aChunk);
-            if (ptr.aboveOrEqual(dataStart) && ptr.belowThan(dataEnd)) {
-                log.string("points into an aligned TLAB chunk of thread ").zhex(thread);
-                return true;
-            }
-            aChunk = HeapChunk.getNext(aChunk);
-        }
-
-        UnalignedHeader uChunk = tlab.getUnalignedChunk();
-        while (uChunk.isNonNull()) {
-            Pointer dataStart = UnalignedHeapChunk.getObjectStart(uChunk);
-            Pointer dataEnd = UnalignedHeapChunk.getObjectEnd(uChunk);
-            if (ptr.aboveOrEqual(dataStart) && ptr.belowThan(dataEnd)) {
-                log.string("points into an unaligned TLAB chunk of thread ").zhex(thread);
-                return true;
-            }
-            uChunk = HeapChunk.getNext(uChunk);
-        }
-
-        return false;
-    }
-
-    @Uninterruptible(reason = "This whole method is unsafe, so it is only uninterruptible to satisfy the checks.")
-    private static Descriptor getTlabUnsafe(IsolateThread thread) {
-        assert SubstrateDiagnostics.isFatalErrorHandlingThread() : "can cause crashes, so it may only be used while printing diagnostics";
-        return ThreadLocalAllocation.getTlab(thread);
-    }
-
     @Override
     @Uninterruptible(reason = "Called during early startup.")
     public boolean verifyImageHeapMapping() {
         for (ImageHeapInfo info : HeapImpl.getImageHeapInfos()) {
             /* Read & write some data at the beginning and end of each writable chunk. */
-            writeToEachChunk(info.getFirstWritableAlignedChunk(), WordFactory.nullPointer());
+            writeToEachChunk(info.getFirstWritableAlignedChunk(), Word.nullPointer());
             writeToEachChunk(info.getFirstWritableUnalignedChunk(), info.getLastWritableUnalignedChunk());
         }
         return true;
@@ -873,7 +862,7 @@ public final class HeapImpl extends Heap {
         }
     }
 
-    private static class DumpHeapSettingsAndStatistics extends DiagnosticThunk {
+    private static final class DumpHeapSettingsAndStatistics extends DiagnosticThunk {
         @Override
         public int maxInvocationCount() {
             return 1;
@@ -883,10 +872,7 @@ public final class HeapImpl extends Heap {
         @RestrictHeapAccess(access = RestrictHeapAccess.Access.NO_ALLOCATION, reason = "Must not allocate while printing diagnostics.")
         public void printDiagnostics(Log log, ErrorContext context, int maxDiagnosticLevel, int invocationCount) {
             log.string("Heap settings and statistics:").indent(true);
-            log.string("Supports isolates: ").bool(SubstrateOptions.SpawnIsolates.getValue()).newline();
-            log.string("Heap base: ").zhex(KnownIntrinsics.heapBase()).newline();
-            log.string("Object reference size: ").signed(ConfigurationValues.getObjectLayout().getReferenceSize()).newline();
-            log.string("Reserved object header bits: 0b").number(Heap.getHeap().getObjectHeader().getReservedBitsMask(), 2, false).newline();
+            log.string("Reserved hub pointer bits: 0b").number(Heap.getHeap().getObjectHeader().getReservedHubBitsMask(), 2, false).newline();
 
             log.string("Aligned chunk size: ").unsigned(HeapParameters.getAlignedHeapChunkSize()).newline();
             log.string("Large array threshold: ").unsigned(HeapParameters.getLargeArrayThreshold()).newline();
@@ -901,7 +887,7 @@ public final class HeapImpl extends Heap {
         }
     }
 
-    private static class DumpHeapUsage extends DiagnosticThunk {
+    private static final class DumpHeapUsage extends DiagnosticThunk {
         @Override
         public int maxInvocationCount() {
             return 1;
@@ -916,7 +902,7 @@ public final class HeapImpl extends Heap {
         }
     }
 
-    private static class DumpChunkInformation extends DiagnosticThunk {
+    private static final class DumpGCPolicy extends DiagnosticThunk {
         @Override
         public int maxInvocationCount() {
             return 1;
@@ -925,14 +911,162 @@ public final class HeapImpl extends Heap {
         @Override
         @RestrictHeapAccess(access = RestrictHeapAccess.Access.NO_ALLOCATION, reason = "Must not allocate while printing diagnostics.")
         public void printDiagnostics(Log log, ErrorContext context, int maxDiagnosticLevel, int invocationCount) {
-            HeapImpl heap = HeapImpl.getHeapImpl();
-            heap.logImageHeapPartitionBoundaries(log);
+            CollectionPolicy policy = GCImpl.getPolicy();
 
-            log.string("Heap chunks: E=eden, S=survivor, O=old, F=free; A=aligned chunk, U=unaligned chunk; T=to space").indent(true);
-            heap.logChunks(log);
+            log.string("GC policy:").indent(true);
+            log.string("Name: ").string(policy.getName()).newline();
+            log.string("Max eden size: ").unsigned(policy.getMaximumEdenSize()).newline();
+            log.string("Max survivor size: ").unsigned(policy.getMaximumSurvivorSize()).newline();
+            log.string("Max young size: ").unsigned(policy.getMaximumYoungGenerationSize()).newline();
+            log.string("Max old size: ").unsigned(policy.getMaximumOldSize()).newline();
+            log.string("Max heap size: ").unsigned(policy.getMaximumHeapSize()).newline();
             log.indent(false);
         }
     }
+
+    private static final class DumpImageHeapInfo extends DiagnosticThunk {
+        @Override
+        public int maxInvocationCount() {
+            return 1;
+        }
+
+        @Override
+        @RestrictHeapAccess(access = RestrictHeapAccess.Access.NO_ALLOCATION, reason = "Must not allocate while printing diagnostics.")
+        public void printDiagnostics(Log log, ErrorContext context, int maxDiagnosticLevel, int invocationCount) {
+            log.string("Image heap boundaries:").indent(true);
+            for (ImageHeapInfo info : HeapImpl.getImageHeapInfos()) {
+                info.print(log);
+            }
+            log.indent(false);
+
+            if (AuxiliaryImageHeap.isPresent()) {
+                ImageHeapInfo auxHeapInfo = AuxiliaryImageHeap.singleton().getImageHeapInfo();
+                if (auxHeapInfo != null) {
+                    log.string("Auxiliary image heap boundaries:").indent(true);
+                    auxHeapInfo.print(log);
+                    log.indent(false);
+                }
+            }
+        }
+    }
+
+    private static final class DumpChunkInfo extends DiagnosticThunk {
+        @Override
+        public int maxInvocationCount() {
+            return 2;
+        }
+
+        @Override
+        @RestrictHeapAccess(access = RestrictHeapAccess.Access.NO_ALLOCATION, reason = "Must not allocate while printing diagnostics.")
+        public void printDiagnostics(Log log, ErrorContext context, int maxDiagnosticLevel, int invocationCount) {
+            log.string("Heap chunks: E=eden, S=survivor, O=old, F=free; A=aligned chunk, U=unaligned chunk; T=to space").indent(true);
+            boolean allowUnsafe = invocationCount == 1 && SubstrateDiagnostics.DiagnosticLevel.unsafeOperationsAllowed(maxDiagnosticLevel);
+            HeapImpl.getHeapImpl().logChunks(log, allowUnsafe);
+            log.indent(false);
+        }
+    }
+
+    private static final class ImageHeapChunkLogger implements HeapChunkVisitor {
+        private Log log;
+
+        @SuppressWarnings("hiding")
+        void initialize(Log log) {
+            this.log = log;
+        }
+
+        @Override
+        public void visitAlignedChunk(AlignedHeader chunk) {
+            Pointer bottom = AlignedHeapChunk.getObjectsStart(chunk);
+            Pointer top = HeapChunk.getTopPointer(chunk);
+            Pointer end = AlignedHeapChunk.getObjectsEnd(chunk);
+            HeapChunkLogging.logChunk(log, chunk, bottom, top, end, true, "I", false);
+        }
+
+        @Override
+        public void visitUnalignedChunk(UnalignedHeader chunk) {
+            Pointer bottom = UnalignedHeapChunk.getObjectStart(chunk);
+            Pointer top = HeapChunk.getTopPointer(chunk);
+            Pointer end = UnalignedHeapChunk.getObjectEnd(chunk);
+            HeapChunkLogging.logChunk(log, chunk, bottom, top, end, false, "I", false);
+        }
+    }
+
+    public static DynamicHub allocateDynamicHub(int vTableSlots) {
+        AllocateDynamicHubOp vmOp = new AllocateDynamicHubOp(vTableSlots);
+        vmOp.enqueue();
+        return vmOp.result;
+    }
+
+    private static class AllocateDynamicHubOp extends JavaVMOperation {
+        int vTableSlots;
+        DynamicHub result;
+
+        AllocateDynamicHubOp(int vTableSlots) {
+            super(VMOperationInfos.get(AllocateDynamicHubOp.class, "Allocate DynamicHub", SystemEffect.SAFEPOINT));
+            this.vTableSlots = vTableSlots;
+        }
+
+        @Override
+        @Uninterruptible(reason = "Called from uninterruptible code.", mayBeInlined = true)
+        public boolean isGC() {
+            /* needs to append chunks into oldGen */
+            return true;
+        }
+
+        @Override
+        protected void operate() {
+            DynamicHub hubOfDynamicHub = DynamicHub.fromClass(Class.class);
+            /*
+             * Note that layoutEncoding already encodes the size of a DynamicHub and it is aware of
+             * its hybrid nature, including the size required for a VTable slot.
+             *
+             * Also note that inlined fields like `closedTypeWorldTypeCheckSlots` are not relevant
+             * here, as they are not available in the open type world configuration.
+             */
+            UnsignedWord size = LayoutEncoding.getArrayAllocationSize(hubOfDynamicHub.getLayoutEncoding(), vTableSlots);
+
+            Pointer memory = Word.nullPointer();
+            if (getHeapImpl().lastDynamicHubChunk.isNonNull()) {
+                /*
+                 * GR-57355: move this fast-path out of vmOp. Needs some locking (it's not
+                 * thread-local)
+                 */
+                memory = AlignedHeapChunk.allocateMemory(getHeapImpl().lastDynamicHubChunk, size);
+            }
+
+            if (memory.isNull()) {
+                /* Either no storage for DynamicHubs yet or we are out of memory */
+                allocateNewDynamicHubChunk();
+
+                memory = AlignedHeapChunk.allocateMemory(getHeapImpl().lastDynamicHubChunk, size);
+            }
+
+            VMError.guarantee(memory.isNonNull(), "failed to allocate DynamicHub");
+
+            /* DynamicHubs live allocated on aligned heap chunks */
+            boolean unaligned = false;
+            result = (DynamicHub) FormatArrayNode.formatArray(memory, DynamicHub.class, vTableSlots, true, unaligned, AllocationSnippets.FillContent.WITH_ZEROES, true);
+        }
+
+        private static void allocateNewDynamicHubChunk() {
+            /*
+             * GR-60085: Should be a dedicated generation. Make sure that those chunks are close to
+             * the heap base. The hub is stored as offset relative to the heap base. There are 5
+             * status bits in the header and in addition, compressed references use a three-bit
+             * shift that word-aligns objects. This results in a 35-bit address range of 32 GB, of
+             * which DynamicHubs must reside in the lowest 1 GB.
+             */
+            OldGeneration oldGeneration = getHeapImpl().getOldGeneration();
+
+            /*
+             * GR-60085: DynamicHub objects must never be be moved. Pin them either by (1) pinning
+             * each DynamicHub, or (2) mark the whole chunk as pinned (not supported yet).
+             */
+            getHeapImpl().lastDynamicHubChunk = oldGeneration.requestAlignedChunk();
+
+            oldGeneration.appendChunk(getHeapImpl().lastDynamicHubChunk);
+        }
+    }
 }
 
 @TargetClass(value = java.lang.Runtime.class, onlyWith = UseSerialOrEpsilonGC.class)
