@@ -26,11 +26,6 @@ package com.oracle.svm.core.heap.dump;
 
 import static com.oracle.svm.core.heap.RestrictHeapAccess.Access.NO_ALLOCATION;
 
-import org.graalvm.compiler.api.replacements.Fold;
-import org.graalvm.compiler.core.common.NumUtil;
-import org.graalvm.compiler.nodes.java.ArrayLengthNode;
-import org.graalvm.compiler.word.ObjectAccess;
-import org.graalvm.compiler.word.Word;
 import org.graalvm.nativeimage.CurrentIsolate;
 import org.graalvm.nativeimage.IsolateThread;
 import org.graalvm.nativeimage.Platform;
@@ -40,11 +35,9 @@ import org.graalvm.nativeimage.c.function.CodePointer;
 import org.graalvm.word.Pointer;
 import org.graalvm.word.UnsignedWord;
 import org.graalvm.word.WordBase;
-import org.graalvm.word.WordFactory;
 
 import com.oracle.svm.core.NeverInline;
 import com.oracle.svm.core.StaticFieldsSupport;
-import com.oracle.svm.core.SubstrateOptions;
 import com.oracle.svm.core.c.NonmovableArray;
 import com.oracle.svm.core.code.CodeInfo;
 import com.oracle.svm.core.code.CodeInfoAccess;
@@ -60,6 +53,8 @@ import com.oracle.svm.core.collections.GrowableWordArrayAccess;
 import com.oracle.svm.core.config.ConfigurationValues;
 import com.oracle.svm.core.deopt.DeoptimizedFrame;
 import com.oracle.svm.core.heap.CodeReferenceMapDecoder;
+import com.oracle.svm.core.heap.FillerArray;
+import com.oracle.svm.core.heap.FillerObject;
 import com.oracle.svm.core.heap.Heap;
 import com.oracle.svm.core.heap.NoAllocationVerifier;
 import com.oracle.svm.core.heap.ObjectReferenceVisitor;
@@ -78,7 +73,9 @@ import com.oracle.svm.core.hub.DynamicHub;
 import com.oracle.svm.core.hub.LayoutEncoding;
 import com.oracle.svm.core.jdk.UninterruptibleUtils.CharReplacer;
 import com.oracle.svm.core.jdk.UninterruptibleUtils.ReplaceDotWithSlash;
+import com.oracle.svm.core.layeredimagesingleton.MultiLayeredImageSingleton;
 import com.oracle.svm.core.log.Log;
+import com.oracle.svm.core.nmt.NmtCategory;
 import com.oracle.svm.core.os.BufferedFileOperationSupport;
 import com.oracle.svm.core.os.BufferedFileOperationSupport.BufferedFile;
 import com.oracle.svm.core.os.RawFileOperationSupport.RawFileDescriptor;
@@ -86,12 +83,19 @@ import com.oracle.svm.core.snippets.KnownIntrinsics;
 import com.oracle.svm.core.stack.JavaStackWalker;
 import com.oracle.svm.core.stack.StackFrameVisitor;
 import com.oracle.svm.core.thread.PlatformThreads;
-import com.oracle.svm.core.thread.ThreadingSupportImpl;
+import com.oracle.svm.core.thread.RecurringCallbackSupport;
 import com.oracle.svm.core.thread.VMOperation;
 import com.oracle.svm.core.thread.VMThreads;
-import com.oracle.svm.core.threadlocal.VMThreadLocalMTSupport;
+import com.oracle.svm.core.threadlocal.VMThreadLocalSupport;
+import com.oracle.svm.core.util.TimeUtils;
 import com.oracle.svm.core.util.VMError;
 
+import jdk.graal.compiler.api.replacements.Fold;
+import jdk.graal.compiler.core.common.NumUtil;
+import jdk.graal.compiler.nodes.java.ArrayLengthNode;
+import jdk.graal.compiler.word.ObjectAccess;
+import jdk.graal.compiler.word.Word;
+
 /**
  * This class dumps the image heap and the Java heap into a file (HPROF binary format), similar to
  * the HotSpot classes in {@code heapDumper.cpp}. The heap dumping needs additional metadata that is
@@ -417,7 +421,7 @@ public class HeapDumpWriter {
 
     public boolean dumpHeap(RawFileDescriptor fd) {
         assert VMOperation.isInProgressAtSafepoint();
-        assert ThreadingSupportImpl.isRecurringCallbackPaused();
+        assert RecurringCallbackSupport.isCallbackUnsupportedOrTimerSuspended();
 
         noAllocationVerifier.open();
         try {
@@ -446,7 +450,7 @@ public class HeapDumpWriter {
     private boolean initialize(RawFileDescriptor fd) {
         assert topLevelRecordBegin == -1 && subRecordBegin == -1 && !error;
 
-        this.f = file().allocate(fd);
+        this.f = file().allocate(fd, NmtCategory.HeapDump);
         if (f.isNull()) {
             return false;
         }
@@ -458,7 +462,7 @@ public class HeapDumpWriter {
 
         assert f.isNull() || error || file().getUnflushedDataSize(f) == 0;
         file().free(f);
-        this.f = WordFactory.nullPointer();
+        this.f = Word.nullPointer();
 
         this.topLevelRecordBegin = -1;
         this.subRecordBegin = -1;
@@ -504,7 +508,7 @@ public class HeapDumpWriter {
         writeUTF8("JAVA PROFILE 1.0.2");
         writeByte((byte) 0);
         writeInt(wordSize());
-        writeLong(System.currentTimeMillis());
+        writeLong(TimeUtils.currentTimeMillis());
     }
 
     private void startTopLevelRecord(HProfTopLevelRecord tag) {
@@ -565,7 +569,7 @@ public class HeapDumpWriter {
     private void writeSymbol(FieldName fieldName) {
         startTopLevelRecord(HProfTopLevelRecord.UTF8);
         writeFieldNameId(fieldName);
-        write((Pointer) FieldNameAccess.getChars(fieldName), WordFactory.unsigned(FieldNameAccess.getLength(fieldName)));
+        write((Pointer) FieldNameAccess.getChars(fieldName), Word.unsigned(FieldNameAccess.getLength(fieldName)));
         endTopLevelRecord();
     }
 
@@ -690,9 +694,9 @@ public class HeapDumpWriter {
 
     private static Object getStaticFieldDataHolder(HProfType type) {
         if (type == HProfType.NORMAL_OBJECT) {
-            return StaticFieldsSupport.getStaticObjectFields();
+            return StaticFieldsSupport.getStaticObjectFieldsAtRuntime(MultiLayeredImageSingleton.UNKNOWN_LAYER_NUMBER);
         } else {
-            return StaticFieldsSupport.getStaticPrimitiveFields();
+            return StaticFieldsSupport.getStaticPrimitiveFieldsAtRuntime(MultiLayeredImageSingleton.UNKNOWN_LAYER_NUMBER);
         }
     }
 
@@ -742,10 +746,8 @@ public class HeapDumpWriter {
     }
 
     private void writeThreadLocals(IsolateThread isolateThread, int threadSerialNum) {
-        if (SubstrateOptions.MultiThreaded.getValue()) {
-            threadLocalsVisitor.initialize(threadSerialNum);
-            VMThreadLocalMTSupport.singleton().walk(isolateThread, threadLocalsVisitor);
-        }
+        threadLocalsVisitor.initialize(threadSerialNum);
+        VMThreadLocalSupport.singleton().walk(isolateThread, threadLocalsVisitor);
     }
 
     private void writeJNIGlobals() {
@@ -779,7 +781,7 @@ public class HeapDumpWriter {
             writeLargeObjects(largeObjects);
         } finally {
             GrowableWordArrayAccess.freeData(largeObjects);
-            largeObjects = WordFactory.nullPointer();
+            largeObjects = Word.nullPointer();
         }
     }
 
@@ -973,8 +975,8 @@ public class HeapDumpWriter {
      */
     private static int calculateMaxArrayLength(Object array, int elementSize, int recordHeaderSize) {
         int length = ArrayLengthNode.arrayLength(array);
-        UnsignedWord lengthInBytes = WordFactory.unsigned(length).multiply(elementSize);
-        UnsignedWord maxBytes = WordFactory.unsigned(MAX_UNSIGNED_INT).subtract(recordHeaderSize);
+        UnsignedWord lengthInBytes = Word.unsigned(length).multiply(elementSize);
+        UnsignedWord maxBytes = Word.unsigned(MAX_UNSIGNED_INT).subtract(recordHeaderSize);
 
         if (lengthInBytes.belowOrEqual(maxBytes)) {
             return length;
@@ -998,7 +1000,7 @@ public class HeapDumpWriter {
 
     private void writeU1ArrayData(Object array, int length, int arrayBaseOffset) {
         Pointer data = getArrayData(array, arrayBaseOffset);
-        write(data, WordFactory.unsigned(length));
+        write(data, Word.unsigned(length));
     }
 
     private void writeU2ArrayData(Object array, int length, int arrayBaseOffset) {
@@ -1080,12 +1082,12 @@ public class HeapDumpWriter {
         /*
          * HotSpot writes Class objects only as GC_CLASS_DUMP and never as GC_INSTANCE_DUMP records.
          * It also always uses the address of the mirror class as the class id. This has the effect
-         * that the heap dump only contains a very limited set of information for class objects.
+         * that the heap dump only contains very limited information about Class objects.
          *
-         * It is handy to have detailed information about the DynamicHub in the heap dump. Ideally,
-         * we would just write both a GC_CLASS_DUMP and a GC_INSTANCE_DUMP record with the same id
-         * but that breaks VisualVM in a weird way. Therefore, we are using different ids for the
-         * GC_CLASS_DUMP and GC_INSTANCE_DUMP records.
+         * For Native Image, we choose a different approach because it is handy to have detailed
+         * information about the DynamicHub in the heap dump. Ideally, we would just write both a
+         * GC_CLASS_DUMP and a GC_INSTANCE_DUMP record with the same id but that breaks VisualVM in
+         * a weird way. So, we generate an artificial id for GC_CLASS_DUMP entries.
          */
         Word hubAddress = Word.objectToUntrackedPointer(hub);
         if (hubAddress.isNonNull()) {
@@ -1205,30 +1207,32 @@ public class HeapDumpWriter {
 
         @Override
         @RestrictHeapAccess(access = NO_ALLOCATION, reason = "Heap dumping must not allocate.")
-        protected boolean visitFrame(Pointer sp, CodePointer ip, CodeInfo codeInfo, DeoptimizedFrame deoptimizedFrame) {
-            if (deoptimizedFrame != null) {
-                markAsGCRoot(deoptimizedFrame);
+        protected boolean visitRegularFrame(Pointer sp, CodePointer ip, CodeInfo codeInfo) {
+            /*
+             * All references that are on the stack need to be marked as GC roots. Our information
+             * is not necessarily precise enough to identify the exact Java-level stack frame to
+             * which a reference belongs. Therefore, we just dump the data in a way that it gets
+             * associated with the deepest inlined Java-level stack frame of each compilation unit.
+             */
+            markStackValuesAsGCRoots(sp, ip, codeInfo);
 
-                for (DeoptimizedFrame.VirtualFrame frame = deoptimizedFrame.getTopFrame(); frame != null; frame = frame.getCaller()) {
-                    visitFrame(frame.getFrameInfo());
-                    nextFrameId++;
-                }
-            } else {
-                /*
-                 * All references that are on the stack need to be marked as GC roots. Our
-                 * information is not necessarily precise enough to identify the exact Java-level
-                 * stack frame to which a reference belongs. Therefore, we just dump the data in a
-                 * way that it gets associated with the deepest inlined Java-level stack frame of
-                 * each compilation unit.
-                 */
-                markStackValuesAsGCRoots(sp, ip, codeInfo);
+            frameInfoCursor.initialize(codeInfo, ip, true);
+            while (frameInfoCursor.advance()) {
+                FrameInfoQueryResult frame = frameInfoCursor.get();
+                visitFrame(frame);
+                nextFrameId++;
+            }
+            return true;
+        }
 
-                frameInfoCursor.initialize(codeInfo, ip, true);
-                while (frameInfoCursor.advance()) {
-                    FrameInfoQueryResult frame = frameInfoCursor.get();
-                    visitFrame(frame);
-                    nextFrameId++;
-                }
+        @Override
+        @RestrictHeapAccess(access = NO_ALLOCATION, reason = "Heap dumping must not allocate.")
+        protected boolean visitDeoptimizedFrame(Pointer originalSP, CodePointer deoptStubIP, DeoptimizedFrame deoptimizedFrame) {
+            markAsGCRoot(deoptimizedFrame);
+
+            for (DeoptimizedFrame.VirtualFrame frame = deoptimizedFrame.getTopFrame(); frame != null; frame = frame.getCaller()) {
+                visitFrame(frame.getFrameInfo());
+                nextFrameId++;
             }
             return true;
         }
@@ -1242,12 +1246,12 @@ public class HeapDumpWriter {
         private void markStackValuesAsGCRoots(Pointer sp, CodePointer ip, CodeInfo codeInfo) {
             if (markGCRoots) {
                 SimpleCodeInfoQueryResult queryResult = StackValue.get(SimpleCodeInfoQueryResult.class);
-                CodeInfoAccess.lookupCodeInfo(codeInfo, CodeInfoAccess.relativeIP(codeInfo, ip), queryResult);
+                CodeInfoAccess.lookupCodeInfo(codeInfo, ip, queryResult);
 
                 NonmovableArray<Byte> referenceMapEncoding = CodeInfoAccess.getStackReferenceMapEncoding(codeInfo);
                 long referenceMapIndex = queryResult.getReferenceMapIndex();
                 if (referenceMapIndex == ReferenceMapIndex.NO_REFERENCE_MAP) {
-                    throw CodeInfoTable.reportNoReferenceMap(sp, ip, codeInfo);
+                    throw CodeInfoTable.fatalErrorNoReferenceMap(sp, ip, codeInfo);
                 }
                 CodeReferenceMapDecoder.walkOffsetsFromPointer(sp, referenceMapEncoding, referenceMapIndex, this, null);
             }
@@ -1255,7 +1259,17 @@ public class HeapDumpWriter {
 
         @Override
         @RestrictHeapAccess(access = NO_ALLOCATION, reason = "Heap dumping must not allocate.")
-        public boolean visitObjectReference(Pointer objRef, boolean compressed, Object holderObject) {
+        public void visitObjectReferences(Pointer firstObjRef, boolean compressed, int referenceSize, Object holderObject, int count) {
+            Pointer pos = firstObjRef;
+            Pointer end = firstObjRef.add(Word.unsigned(count).multiply(referenceSize));
+            while (pos.belowThan(end)) {
+                visitObjectReference(pos, compressed);
+                pos = pos.add(referenceSize);
+            }
+        }
+
+        /** Derived references are not relevant for heap dumping, so we ignore innerOffset. */
+        private void visitObjectReference(Pointer objRef, boolean compressed) {
             assert markGCRoots;
 
             Object obj = ReferenceAccess.singleton().readObjectAt(objRef, compressed);
@@ -1267,7 +1281,6 @@ public class HeapDumpWriter {
                 /* Position of the stack frame in the stack trace. */
                 writeInt(getWrittenFrames());
             }
-            return true;
         }
 
         private void visitFrame(FrameInfoQueryResult frame) {
@@ -1351,16 +1364,22 @@ public class HeapDumpWriter {
 
         @Override
         @RestrictHeapAccess(access = NO_ALLOCATION, reason = "Heap dumping must not allocate.")
-        public boolean visitObject(Object obj) {
-            if (isLarge(obj)) {
-                boolean added = GrowableWordArrayAccess.add(largeObjects, Word.objectToUntrackedPointer(obj));
-                if (!added) {
-                    Log.log().string("Failed to add an element to the large object list. Heap dump will be incomplete.").newline();
+        public void visitObject(Object obj) {
+            if (!isFillerObject(obj)) {
+                if (isLarge(obj)) {
+                    boolean added = GrowableWordArrayAccess.add(largeObjects, Word.objectToUntrackedPointer(obj), NmtCategory.HeapDump);
+                    if (!added) {
+                        Log.log().string("Failed to add an element to the large object list. Heap dump will be incomplete.").newline();
+                    }
+                } else {
+                    writeObject(obj);
                 }
-            } else {
-                writeObject(obj);
             }
-            return true;
+        }
+
+        private static boolean isFillerObject(Object obj) {
+            /* Filler objects increase the size of the heap dump but don't add much value. */
+            return obj.getClass() == FillerArray.class || obj.getClass() == FillerObject.class;
         }
 
         private boolean isLarge(Object obj) {
@@ -1377,10 +1396,10 @@ public class HeapDumpWriter {
                     elementSize = wordSize();
                 }
                 int length = ArrayLengthNode.arrayLength(obj);
-                return WordFactory.unsigned(length).multiply(elementSize);
+                return Word.unsigned(length).multiply(elementSize);
             } else {
                 ClassInfo classInfo = metadata.getClassInfo(obj.getClass());
-                return WordFactory.unsigned(classInfo.getInstanceFieldsDumpSize());
+                return Word.unsigned(classInfo.getInstanceFieldsDumpSize());
             }
         }
     }
@@ -1391,19 +1410,27 @@ public class HeapDumpWriter {
         }
 
         @Override
-        public boolean visitCode(CodeInfo info) {
+        @RestrictHeapAccess(access = NO_ALLOCATION, reason = "Heap dumping must not allocate.")
+        public void visitCode(CodeInfo info) {
             RuntimeCodeInfoAccess.walkObjectFields(info, this);
-            return true;
         }
 
         @Override
         @RestrictHeapAccess(access = NO_ALLOCATION, reason = "Heap dumping must not allocate.")
-        public boolean visitObjectReference(Pointer objRef, boolean compressed, Object holderObject) {
+        public void visitObjectReferences(Pointer firstObjRef, boolean compressed, int referenceSize, Object holderObject, int count) {
+            Pointer pos = firstObjRef;
+            Pointer end = firstObjRef.add(Word.unsigned(count).multiply(referenceSize));
+            while (pos.belowThan(end)) {
+                visitObjectReference(pos, compressed);
+                pos = pos.add(referenceSize);
+            }
+        }
+
+        private void visitObjectReference(Pointer objRef, boolean compressed) {
             Object obj = ReferenceAccess.singleton().readObjectAt(objRef, compressed);
             if (obj != null) {
                 markAsJniGlobalGCRoot(obj);
             }
-            return true;
         }
     }
 
@@ -1421,12 +1448,20 @@ public class HeapDumpWriter {
 
         @Override
         @RestrictHeapAccess(access = NO_ALLOCATION, reason = "Heap dumping must not allocate.")
-        public boolean visitObjectReference(Pointer objRef, boolean compressed, Object holderObject) {
+        public void visitObjectReferences(Pointer firstObjRef, boolean compressed, int referenceSize, Object holderObject, int count) {
+            Pointer pos = firstObjRef;
+            Pointer end = firstObjRef.add(Word.unsigned(count).multiply(referenceSize));
+            while (pos.belowThan(end)) {
+                visitObjectReference(pos, compressed);
+                pos = pos.add(referenceSize);
+            }
+        }
+
+        private void visitObjectReference(Pointer objRef, boolean compressed) {
             Object obj = ReferenceAccess.singleton().readObjectAt(objRef, compressed);
             if (obj != null) {
                 markThreadLocalAsGCRoot(obj);
             }
-            return true;
         }
 
         private void markThreadLocalAsGCRoot(Object obj) {
@@ -1439,6 +1474,6 @@ public class HeapDumpWriter {
         }
     }
 
-    private static class UnknownClass {
+    private static final class UnknownClass {
     }
 }
