@@ -35,7 +35,6 @@ import com.oracle.svm.core.annotate.RecomputeFieldValue;
 import com.oracle.svm.core.annotate.RecomputeFieldValue.Kind;
 import com.oracle.svm.core.annotate.Substitute;
 import com.oracle.svm.core.annotate.TargetClass;
-import com.oracle.svm.core.heap.PhysicalMemory;
 import com.oracle.svm.core.snippets.KnownIntrinsics;
 
 @TargetClass(className = "jdk.internal.misc.VM")
@@ -67,12 +66,10 @@ public final class Target_jdk_internal_misc_VM {
     @Alias @InjectAccessors(DirectMemoryAccessors.class) //
     private static long directMemory;
     @Alias @InjectAccessors(PageAlignDirectMemoryAccessors.class) //
-    private static boolean pageAlignDirectMemory;
+    private static Boolean pageAlignDirectMemory;
 }
 
 final class DirectMemoryAccessors {
-    private static final long DIRECT_MEMORY_DURING_INITIALIZATION = 25 * 1024 * 1024;
-
     /**
      * This field needs to be volatile to ensure that reads emit a LOAD-LOAD barrier. Without this
      * barrier, subsequent reads could be reordered before the read of {@link #initialized},
@@ -103,15 +100,6 @@ final class DirectMemoryAccessors {
              * No value explicitly specified. The default in the JDK in this case is the maximum
              * heap size.
              */
-            if (PhysicalMemory.isInitializationInProgress()) {
-                /*
-                 * When initializing PhysicalMemory, we use NIO/cgroups code that calls
-                 * VM.getDirectMemory(). When this initialization is in progress, we need to prevent
-                 * that Runtime.maxMemory() is called below because it would trigger a recursive
-                 * initialization of PhysicalMemory. So, we return a temporary value.
-                 */
-                return DIRECT_MEMORY_DURING_INITIALIZATION;
-            }
             newDirectMemory = Runtime.getRuntime().maxMemory();
         }
 
@@ -127,7 +115,7 @@ final class PageAlignDirectMemoryAccessors {
     private static volatile boolean initialized;
     private static boolean pageAlignDirectMemory;
 
-    static boolean getPageAlignDirectMemory() {
+    static Boolean getPageAlignDirectMemory() {
         if (!initialized) {
             initialize();
         }
