@@ -26,7 +26,6 @@ package com.oracle.svm.core.jdk;
 
 import static com.oracle.svm.core.heap.RestrictHeapAccess.Access.NO_ALLOCATION;
 
-import org.graalvm.compiler.nodes.UnreachableNode;
 import org.graalvm.nativeimage.ImageSingletons;
 import org.graalvm.nativeimage.LogHandler;
 import org.graalvm.nativeimage.Platforms;
@@ -45,6 +44,8 @@ import com.oracle.svm.core.stack.StackOverflowCheck;
 import com.oracle.svm.core.thread.VMThreads.SafepointBehavior;
 import com.oracle.svm.core.util.VMError;
 
+import jdk.graal.compiler.nodes.UnreachableNode;
+
 @TargetClass(com.oracle.svm.core.util.VMError.class)
 @Platforms(InternalPlatform.NATIVE_ONLY.class)
 final class Target_com_oracle_svm_core_util_VMError {
@@ -161,9 +162,8 @@ public class VMErrorSubstitutions {
                 }
                 if (ex != null) {
                     log.string(": ").exception(ex);
-                } else {
-                    log.newline();
                 }
+                log.newline();
 
                 SubstrateDiagnostics.printFatalError(log, KnownIntrinsics.readCallerStackPointer(), KnownIntrinsics.readReturnAddress());
 
