@@ -24,6 +24,7 @@
  */
 package com.oracle.svm.core.jfr;
 
+import jdk.graal.compiler.word.Word;
 import org.graalvm.nativeimage.CurrentIsolate;
 import org.graalvm.nativeimage.Platform;
 import org.graalvm.nativeimage.Platforms;
@@ -35,7 +36,6 @@ import org.graalvm.nativeimage.c.struct.SizeOf;
 import org.graalvm.nativeimage.c.type.CIntPointer;
 import org.graalvm.word.Pointer;
 import org.graalvm.word.UnsignedWord;
-import org.graalvm.word.WordFactory;
 
 import com.oracle.svm.core.FrameAccess;
 import com.oracle.svm.core.NeverInline;
@@ -211,12 +211,12 @@ public class JfrStackTraceRepository implements JfrRepository {
 
                 /* Hashtable entry allocation failed. */
                 NullableNativeMemory.free(to);
-                to = WordFactory.nullPointer();
+                to = Word.nullPointer();
             }
 
             /* Some allocation failed. */
             statusPtr.write(JfrStackTraceTableEntryStatus.INSERT_FAILED);
-            return WordFactory.nullPointer();
+            return Word.nullPointer();
         }
     }
 
@@ -326,7 +326,7 @@ public class JfrStackTraceRepository implements JfrRepository {
             JfrStackTraceTableEntry entry1 = (JfrStackTraceTableEntry) a;
             JfrStackTraceTableEntry entry2 = (JfrStackTraceTableEntry) b;
             /* We explicitly ignore the field 'serialized' because its value can change. */
-            return entry1.getSize() == entry2.getSize() && LibC.memcmp(entry1.getRawStackTrace(), entry2.getRawStackTrace(), WordFactory.unsigned(entry1.getSize())) == 0;
+            return entry1.getSize() == entry2.getSize() && LibC.memcmp(entry1.getRawStackTrace(), entry2.getRawStackTrace(), Word.unsigned(entry1.getSize())) == 0;
         }
 
         @Override
@@ -385,7 +385,7 @@ public class JfrStackTraceRepository implements JfrRepository {
             table.teardown();
             unflushedEntries = 0;
             JfrBufferAccess.free(buffer);
-            buffer = WordFactory.nullPointer();
+            buffer = Word.nullPointer();
         }
     }
 }
