@@ -31,6 +31,7 @@ import org.graalvm.nativeimage.c.struct.CStruct;
 import org.graalvm.nativeimage.c.type.WordPointer;
 import org.graalvm.word.PointerBase;
 
+import com.oracle.svm.core.jni.headers.JNIFunctionPointerTypes.AllocObjectFunctionPointer;
 import com.oracle.svm.core.jni.headers.JNIFunctionPointerTypes.CallBooleanMethodAFunctionPointer;
 import com.oracle.svm.core.jni.headers.JNIFunctionPointerTypes.CallIntMethodAFunctionPointer;
 import com.oracle.svm.core.jni.headers.JNIFunctionPointerTypes.CallLongMethodAFunctionPointer;
@@ -62,6 +63,7 @@ import com.oracle.svm.core.jni.headers.JNIFunctionPointerTypes.GetSuperclassFunc
 import com.oracle.svm.core.jni.headers.JNIFunctionPointerTypes.IsAssignableFromFunctionPointer;
 import com.oracle.svm.core.jni.headers.JNIFunctionPointerTypes.IsInstanceOfFunctionPointer;
 import com.oracle.svm.core.jni.headers.JNIFunctionPointerTypes.IsSameObjectFunctionPointer;
+import com.oracle.svm.core.jni.headers.JNIFunctionPointerTypes.MonitorEnterExitFunctionPointer;
 import com.oracle.svm.core.jni.headers.JNIFunctionPointerTypes.NewByteArrayFunctionPointer;
 import com.oracle.svm.core.jni.headers.JNIFunctionPointerTypes.NewGlobalRefFunctionPointer;
 import com.oracle.svm.core.jni.headers.JNIFunctionPointerTypes.NewObjectAFunctionPointer;
@@ -231,7 +233,7 @@ public interface JNINativeInterface extends PointerBase {
     void setEnsureLocalCapacity(CFunctionPointer p);
 
     @CField
-    CFunctionPointer getAllocObject();
+    AllocObjectFunctionPointer getAllocObject();
 
     @CField
     void setAllocObject(CFunctionPointer p);
@@ -1371,16 +1373,16 @@ public interface JNINativeInterface extends PointerBase {
     void setUnregisterNatives(CFunctionPointer p);
 
     @CField
-    CFunctionPointer getMonitorEnter();
+    MonitorEnterExitFunctionPointer getMonitorEnter();
 
     @CField
-    void setMonitorEnter(CFunctionPointer p);
+    void setMonitorEnter(MonitorEnterExitFunctionPointer p);
 
     @CField
-    CFunctionPointer getMonitorExit();
+    MonitorEnterExitFunctionPointer getMonitorExit();
 
     @CField
-    void setMonitorExit(CFunctionPointer p);
+    void setMonitorExit(MonitorEnterExitFunctionPointer p);
 
     @CField
     CFunctionPointer getGetJavaVM();
@@ -1472,4 +1474,17 @@ public interface JNINativeInterface extends PointerBase {
 
     @CField
     void setGetModule(CFunctionPointer p);
+
+    // Virtual threads
+
+    @CField
+    void setIsVirtualThread(CFunctionPointer p);
+
+    // JNI_VERSION_24
+
+    @CField
+    CFunctionPointer getGetStringUTFLengthAsLong();
+
+    @CField
+    void setGetStringUTFLengthAsLong(CFunctionPointer p);
 }
