@@ -26,6 +26,9 @@ package com.oracle.svm.core.layeredimagesingleton;
 
 import java.util.EnumSet;
 
+import com.oracle.svm.core.graal.RuntimeCompilation;
+import com.oracle.svm.core.util.VMError;
+
 /**
  * Flags used during build time to determine how the native image generator handles the layered
  * image singleton.
@@ -76,4 +79,22 @@ public enum LayeredImageSingletonBuilderFlags {
 
         return true;
     }
+
+    public static void validateRuntimeLookup(Object singleton) {
+        if (singleton instanceof LayeredImageSingleton layeredSingleton) {
+            if (!layeredSingleton.getImageBuilderFlags().contains(LayeredImageSingletonBuilderFlags.RUNTIME_ACCESS)) {
+                /*
+                 * Runtime compilation installs many singletons into the image which are otherwise
+                 * hosted only. Note the platform checks still apply and can be used to ensure
+                 * certain singleton are not installed into the image.
+                 */
+                if (!RuntimeCompilation.isEnabled()) {
+                    throw VMError.shouldNotReachHere("Layered image singleton without runtime access is in runtime graph: " + layeredSingleton);
+                }
+            }
+            if (layeredSingleton instanceof MultiLayeredImageSingleton) {
+                throw VMError.shouldNotReachHere("Forbidden lookup of MultiLayeredImageSingleton in runtime graph: " + layeredSingleton);
+            }
+        }
+    }
 }
