@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 1999, 2021, Oracle and/or its affiliates. All rights reserved.
+ * Copyright (c) 1999, 2024, Oracle and/or its affiliates. All rights reserved.
  * ORACLE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
  *
  *
@@ -196,6 +196,7 @@ public abstract class X11SurfaceData extends XSurfaceData {
 
     private static Boolean accelerationEnabled = null;
 
+    @Override
     public Raster getRaster(int x, int y, int w, int h) {
         throw new InternalError("not implemented yet");
     }
@@ -272,6 +273,7 @@ public abstract class X11SurfaceData extends XSurfaceData {
         return X11SurfaceDataProxy.createProxy(srcData, graphicsConfig);
     }
 
+    @Override
     public void validatePipe(SunGraphics2D sg2d) {
         if (sg2d.antialiasHint != SunHints.INTVAL_ANTIALIAS_ON &&
             sg2d.paintState <= SunGraphics2D.PAINT_ALPHACOLOR &&
@@ -376,6 +378,7 @@ public abstract class X11SurfaceData extends XSurfaceData {
         }
     }
 
+    @Override
     public RenderLoops getRenderLoops(SunGraphics2D sg2d) {
         if (sg2d.paintState <= SunGraphics2D.PAINT_ALPHACOLOR &&
             sg2d.compositeState <= SunGraphics2D.COMP_ISCOPY)
@@ -385,6 +388,7 @@ public abstract class X11SurfaceData extends XSurfaceData {
         return super.getRenderLoops(sg2d);
     }
 
+    @Override
     public GraphicsConfiguration getDeviceConfiguration() {
         return graphicsConfig;
     }
@@ -458,6 +462,7 @@ public abstract class X11SurfaceData extends XSurfaceData {
      */
     public abstract boolean canSourceSendExposures(int x, int y, int w, int h);
 
+    @Override
     public boolean copyArea(SunGraphics2D sg2d,
                             int x, int y, int w, int h, int dx, int dy)
     {
@@ -591,6 +596,7 @@ public abstract class X11SurfaceData extends XSurfaceData {
         return sType;
     }
 
+    @Override
     public void invalidate() {
         if (isValid()) {
             setInvalid();
@@ -707,10 +713,12 @@ public abstract class X11SurfaceData extends XSurfaceData {
             }
         }
 
+        @Override
         public SurfaceData getReplacement() {
             return peer.getSurfaceData();
         }
 
+        @Override
         public Rectangle getBounds() {
             Rectangle r = peer.getBounds();
             r.x = r.y = 0;
@@ -727,6 +735,7 @@ public abstract class X11SurfaceData extends XSurfaceData {
         /**
          * Returns destination Component associated with this SurfaceData.
          */
+        @Override
         public Object getDestination() {
             return peer.getTarget();
         }
@@ -767,6 +776,7 @@ public abstract class X11SurfaceData extends XSurfaceData {
             makePipes();
         }
 
+        @Override
         public SurfaceData getReplacement() {
             return restoreContents(offscreenImage);
         }
@@ -779,10 +789,12 @@ public abstract class X11SurfaceData extends XSurfaceData {
          * it could choose wrong loop (blit instead of blitbg,
          * for example).
          */
+        @Override
         public int getTransparency() {
             return transparency;
         }
 
+        @Override
         public Rectangle getBounds() {
             return new Rectangle(width, height);
         }
@@ -792,6 +804,7 @@ public abstract class X11SurfaceData extends XSurfaceData {
             return (x < 0 || y < 0 || (x+w) > width || (y+h) > height);
         }
 
+        @Override
         public void flush() {
             /*
              * We need to invalidate the surface before disposing the
@@ -808,6 +821,7 @@ public abstract class X11SurfaceData extends XSurfaceData {
         /**
          * Returns destination Image associated with this SurfaceData.
          */
+        @Override
         public Object getDestination() {
             return offscreenImage;
         }
@@ -825,7 +839,8 @@ public abstract class X11SurfaceData extends XSurfaceData {
 
     private static LazyPipe lazypipe = new LazyPipe();
 
-    public static class LazyPipe extends ValidatePipe {
+    public static final class LazyPipe extends ValidatePipe {
+        @Override
         public boolean validate(SunGraphics2D sg2d) {
             X11SurfaceData xsd = (X11SurfaceData) sg2d.surfaceData;
             if (!xsd.isDrawableValid()) {
