@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2011, 2024, Oracle and/or its affiliates. All rights reserved.
+ * Copyright (c) 2011, 2025, Oracle and/or its affiliates. All rights reserved.
  * ORACLE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
  *
  *
@@ -36,6 +36,7 @@ import java.util.concurrent.atomic.AtomicReference;
 import javax.print.*;
 import javax.print.attribute.PrintRequestAttributeSet;
 import javax.print.attribute.HashPrintRequestAttributeSet;
+import javax.print.attribute.standard.Chromaticity;
 import javax.print.attribute.standard.Copies;
 import javax.print.attribute.standard.Destination;
 import javax.print.attribute.standard.Media;
@@ -73,6 +74,8 @@ public final class CPrinterJob extends RasterPrinterJob {
 
     private Throwable printerAbortExcpn;
 
+    private boolean monochrome = false;
+
     // This is the NSPrintInfo for this PrinterJob. Protect multi thread
     //  access to it. It is used by the pageDialog, jobDialog, and printLoop.
     //  This way the state of these items is shared across these calls.
@@ -101,7 +104,7 @@ public final class CPrinterJob extends RasterPrinterJob {
      * selected by the user.
      * @return {@code true} if the user does not cancel the dialog;
      * {@code false} otherwise.
-     * @exception HeadlessException if GraphicsEnvironment.isHeadless()
+     * @throws HeadlessException if GraphicsEnvironment.isHeadless()
      * returns true.
      * @see java.awt.GraphicsEnvironment#isHeadless
      */
@@ -143,7 +146,7 @@ public final class CPrinterJob extends RasterPrinterJob {
      *            is cancelled; a new {@code PageFormat} object
      *          containing the format indicated by the user if the
      *          dialog is acknowledged.
-     * @exception HeadlessException if GraphicsEnvironment.isHeadless()
+     * @throws HeadlessException if GraphicsEnvironment.isHeadless()
      * returns true.
      * @see java.awt.GraphicsEnvironment#isHeadless
      * @since     1.2
@@ -212,6 +215,11 @@ public final class CPrinterJob extends RasterPrinterJob {
                 setPageRange(-1, -1);
             }
         }
+
+        PrintService service = getPrintService();
+        Chromaticity chromaticity = (Chromaticity)attributes.get(Chromaticity.class);
+        monochrome = chromaticity == Chromaticity.MONOCHROME && service != null &&
+                service.isAttributeCategorySupported(Chromaticity.class);
     }
 
     private void setPageRangeAttribute(int from, int to, boolean isRangeSet) {
@@ -788,6 +796,9 @@ public final class CPrinterJob extends RasterPrinterJob {
                 Graphics2D pathGraphics = new CPrinterGraphics(delegate, printerJob); // Just stores delegate into an ivar
                 Rectangle2D pageFormatArea = getPageFormatArea(page);
                 initPrinterGraphics(pathGraphics, pageFormatArea);
+                if (monochrome) {
+                    pathGraphics = new GrayscaleProxyGraphics2D(pathGraphics, printerJob);
+                }
                 painter.print(pathGraphics, FlipPageFormat.getOriginal(page), pageIndex);
                 delegate.dispose();
                 delegate = null;
