@@ -280,7 +280,14 @@ public class FormatData_cgg extends ListResourceBundle {
             { "japanese.QuarterAbbreviations", metaValue_generic_QuarterAbbreviations },
             { "japanese.TimePatterns", metaValue_generic_TimePatterns },
             { "java.time.japanese.DatePatterns", metaValue_java_time_generic_DatePatterns },
-            { "japanese.DatePatterns", metaValue_generic_DatePatterns },
+            { "japanese.DatePatterns",
+                new String[] {
+                    "EEEE, d MMMM yyyy GGGG",
+                    "d MMMM yyyy GGGG",
+                    "d MMM y GGGG",
+                    "dd/MM/y G",
+                }
+            },
             { "japanese.DateFormatItem.hm", metaValue_generic_DateFormatItem_hm },
             { "japanese.DateFormatItem.yMMMEd", metaValue_generic_DateFormatItem_yMMMEd },
             { "japanese.DateFormatItem.yMEd", metaValue_generic_DateFormatItem_yMEd },
